//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
program lossyWAV;
//==============================================================================
// Conditional Defines
//==============================================================================
{$R *.RES}
//==============================================================================
uses
  nCore in 'units\nCore.pas',
  fftw_interface in 'units\fftw_interface.pas',
  {$IFNDEF USEASM}
  nFFT_pas in 'units\nFFT_pas.pas',
  {$ELSE}
  nFFT_asm in 'units\nFFT_asm.pas',
  {$ENDIF}
  nParameter in 'units\nParameter.pas',
  nFillFFT in 'units\nFillFFT.pas',
  nInitialise in 'units\nInitialise.pas',
  nMaths in 'units\nMaths.pas',
  nOutput in 'units\nOutput.pas',
  nProcess in 'units\nProcess.pas',
  nRemoveBits in 'units\nRemoveBits.pas',
  nSGNS in 'units\nSGNS.pas',
  nShiftBlocks in 'units\nShiftBlocks.pas',
  nSpreading in 'units\nSpreading.pas',
  nWav in 'units\nWav.pas',
  Sysutils,
  Math;

//==============================================================================
begin
//==============================================================================
  Check_Switches();

  if Parameters.merging then
    MergeFiles()
  else
  begin
    if not openWavIO then
      lossyWAVerror('Error initialising wavIO unit.',$11);

    if Global.codec_block_size=0 then
      lossyWAVerror('Error initialising wavIO unit.',$11);

    Initial_Setup();

    if not readNextNextCodecBlock() then
      lossyWAVerror('Error reading from input file.',$21);

    Global.blocks_processed:=0;
    //==========================================================================
    // Main processing loop.
    //==========================================================================
    while (BlockData.Size.next>0) do
    begin
      shift_codec_blocks();

      readNextNextCodecBlock();

      process_this_codec_block();

      if not writeNextBtrdCodecBlock() then
        lossyWAVerror('Error writing to output file.',$21);

      if Parameters.correction and not writeNextCorrCodecBlock() then
        lossyWAVerror('Error writing to correction file.',$22);
    //==========================================================================
    end;
    //==========================================================================

    if not CloseWavIO() then
      lossyWAVerror('Error closing wavIO unit.',$11);

    Write_Cleanup();
  end;
//==============================================================================
end.
//==============================================================================

