{--------------------------------------------------------------}
{ Delphi interface to the FFTW library -- FFTW Version 3.0.1.  }
{ Note that this interface is incomplete. Additional function  }
{ interface entries may be added in an anologous manner, see   }
{ FFTW for more details.                                       }
{                                                              }
{ Last modified 22/DEC/03                                      }
{ Written by Mark G. Beckett (g.beckett@epcc.ed.ac.uk          }
{--------------------------------------------------------------}
{ Modified 07/JUN/2009                                         }
{ by Nick Currie (lossyWAV at yahoo dot com)                   }
{ Now dynamically links rather than statically. Required to    }
{ stop program crashing if libFFTW3-3.DLL is not available.    }
{--------------------------------------------------------------}
{ Modified 14/APR/2010                                         }
{ by Nick Currie (lossyWAV at hotmail <dot> co <dot> uk        }
{--------------------------------------------------------------}

//==============================================================================
Unit FFTW_interface;
//==============================================================================
Interface
 {EPCC (MGB) - Note that the FFTW memory allocation routines can not be used
  for data arrays, because the DLL is allocated a separate heap that is not
  accessible from the Delphi runtime environment.}
//==============================================================================
Uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================
Type
  FFTW_Rec = Record
    Plan_DFT_r2c_1d  : Function (n : Integer; inData : PDouble; outData : PDouble; flags : Longword) : Pointer; cdecl;
    Plan_DFT_1d      : Function (n : Integer; inData : PDouble; outData : PDouble; Sign : Integer; flags : Longword) : Pointer; cdecl;
    Execute          : Procedure(plan : Pointer); cdecl;
    Destroy_Plan     : Procedure(plan : Pointer); cdecl;
    Plans            : Array[0..MAX_FFT_BIT_LENGTH] of pointer;
  End;
//==============================================================================
Const
 {EPCC (MGB) - FFTW documented constants, taken from "api/FFTW-3.h". Comments
  to the right of the definitions are transcribed from the original header
  file.}

  FFTW_FORWARD            = -1;
  FFTW_BACKWARD           = 1;

  FFTW_MEASURE            = 0;
  FFTW_DESTROY_INPUT      = 1 shl 0; {01h; 1U << 0}
  FFTW_UNALIGNED          = 1 shl 1; {02h; 1U << 1}
  FFTW_CONSERVE_MEMORY    = 1 shl 2; {04h; 1U << 2}
  FFTW_EXHAUSTIVE         = 1 shl 3; {08h; 1U << 3} {NO_EXHAUSTIVE is default }
  FFTW_PRESERVE_INPUT     = 1 shl 4; {10h; 1U << 4} {cancels FFTW_DESTROY_INPUT}
  FFTW_PATIENT            = 1 shl 5; {20h; 1U << 5} {IMPATIENT is default }
  FFTW_ESTIMATE           = 1 shl 6; {40h; 1U << 6}

 {FFTW undocumented constants have not been defined in this implementation.
  They are not required for typical usage of the library.}

Var
  FFTW              : FFTW_Rec;
//==============================================================================

Function  CHECK_FFTW_DLL_Loaded : Boolean;

Function  FFTW_Initialised  : Longbool;

//==============================================================================
Implementation
//==============================================================================
Uses
  Windows;
//==============================================================================
Const
  UNITNAME='FFTW_Interface';
Var
  FFTW_DLL_Handle   : THandle;
  FFTW_DLL_Loaded   : LongBool;

  Unit_Data_Packer2 : Array[0..2] of Integer;
//==============================================================================

//==============================================================================
Function Check_Initialised(var FFTW_Record : FFTW_Rec) : LongBool;
//==============================================================================
Begin
//==============================================================================
  with FFTW_Record do
    Check_Initialised:=((@plan_dft_r2c_1d<>Nil) and (@plan_dft_1d<>Nil) and (@execute<>Nil) and (@destroy_plan<>Nil))
//==============================================================================
End;
//==============================================================================

//==============================================================================
Function FFTW_Initialised : Longbool;
//==============================================================================
Begin
//==============================================================================
  FFTW_Initialised:=FFTW_DLL_Loaded and (FFTW_DLL_Handle<>0);
//==============================================================================
End;
//==============================================================================

//==============================================================================
Procedure FFTW_Initialise;
//==============================================================================
Begin
//==============================================================================
  FFTW_DLL_Handle:=LoadLibrary('libFFTW3-3.DLL');
  if (FFTW_DLL_Handle<>0)then
  Begin
    FFTW.plan_dft_r2c_1d  :=GetProcAddress(FFTW_DLL_Handle,'fftw_plan_dft_r2c_1d');
    FFTW.plan_dft_1d      :=GetProcAddress(FFTW_DLL_Handle,'fftw_plan_dft_1d');
    FFTW.execute          :=GetProcAddress(FFTW_DLL_Handle,'fftw_execute');
    FFTW.destroy_plan     :=GetProcAddress(FFTW_DLL_Handle,'fftw_destroy_plan');
  End;
  FFTW_DLL_Loaded:=Check_Initialised(FFTW) and (FFTW_DLL_Handle<>0);
//==============================================================================
End;
//==============================================================================

//==============================================================================
Function CHECK_FFTW_DLL_Loaded : Boolean;
//==============================================================================
Begin
//==============================================================================
  CHECK_FFTW_DLL_Loaded:=FFTW_DLL_Loaded;
//==============================================================================
End;
//==============================================================================

//==============================================================================
Procedure FFTW_Cleanup;
//==============================================================================
var
  fc_i : Integer;
Begin
//==============================================================================
  If (FFTW_DLL_Handle<>0) then
  Begin
    With FFTW do
    Begin
      for fc_i:=1 to MAX_FFT_BIT_LENGTH do
        if plans[fc_i]<>nil then
          DESTROY_PLAN(FFTW.plans[fc_i]);

      plan_dft_r2c_1d :=Nil;
      plan_dft_1d     :=Nil;
      execute         :=Nil;
      destroy_plan    :=Nil;
    End;
    FreeLibrary(FFTW_DLL_Handle);
    FFTW_DLL_Handle:=0;
    FFTW_DLL_Loaded:=False;
  End;
//==============================================================================
End;
//==============================================================================

//==============================================================================
Initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  FFTW_Initialise;
//==============================================================================

//==============================================================================
Finalization
//==============================================================================
  FFTW_Cleanup;
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================

