//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nOutput;
//==============================================================================
interface
//==============================================================================
type
  bit_removal_history_array = array[0..8388607] of byte;
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
  bit_removal_history : ^bit_removal_history_array;
//==============================================================================
procedure write_cleanup;
procedure lsb_analysis;
Procedure Process_Output;
//==============================================================================
implementation
//==============================================================================
uses
  nCore,
  fftw_interface,
  nParameter,
  nMaths,
  nSGNS,
  nSpreading,
  Math,
  SysUtils;
//==============================================================================
const
  UNITNAME            = 'nOutput';

  hyphen_string       : String[255] = '---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------';
  bits_filled         : String[255] = 'OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO000000000000OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO000000000000OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO000000000000OOOOOO';
  bits_empty          : String[255] = '...............................................................................................................................................................................................................................................................';
  space_string        : String[255] = '                                                                                                                                                                                                                                                               ';
  results_offset      : array[0..MAX_CHANNELS-1] of Integer = (0*34*INTEGER_SIZE,1*34*INTEGER_SIZE,2*34*INTEGER_SIZE,3*34*INTEGER_SIZE,4*34*INTEGER_SIZE,5*34*INTEGER_SIZE,6*34*INTEGER_SIZE,7*34*INTEGER_SIZE);
var
  DATA_block_value    : Integer;
  BTRD_block_value    : Integer;
  CORR_block_value    : Integer;

  DATA_block_lsb      : Array[0..MAX_CHANNELS-1,0..33] of integer;
  DATA_block_msb      : Array[0..MAX_CHANNELS-1,0..33] of integer;
  DATA_sample_lsb     : Array[0..MAX_CHANNELS-1,0..33] of integer;
  DATA_sample_msb     : Array[0..MAX_CHANNELS-1,0..33] of integer;

  BTRD_block_lsb      : Array[0..MAX_CHANNELS-1,0..33] of integer;
  BTRD_block_msb      : Array[0..MAX_CHANNELS-1,0..33] of integer;
  BTRD_sample_lsb     : Array[0..MAX_CHANNELS-1,0..33] of integer;
  BTRD_sample_msb     : Array[0..MAX_CHANNELS-1,0..33] of integer;

  CORR_block_lsb      : Array[0..MAX_CHANNELS-1,0..33] of integer;
  CORR_block_msb      : Array[0..MAX_CHANNELS-1,0..33] of integer;
  CORR_sample_lsb     : Array[0..MAX_CHANNELS-1,0..33] of integer;
  CORR_sample_msb     : Array[0..MAX_CHANNELS-1,0..33] of integer;

  head_bar            : String;
  top_bar             : String;
  mid_bar             : String;
  Display_Width       : Integer;
  bar_length          : Integer;
  Titles              : Array[0..15] of String;
  Header              : String;
  Unit_Data_Packer2   : Array[0..0] of Integer;
//==============================================================================

  //============================================================================
  Procedure Make_Bars(MB_Width : Integer; MB_Item : Integer; MB_Value : Integer; MB_Count : Integer; MB_Summary : Integer = 0);
  //============================================================================
  Var
    mb_i : Integer;
  //============================================================================
  Begin
  //============================================================================
    if MB_Summary>0 then
      Bar_Length:=((MB_Width-(MB_Item+2)-(MB_Value+1)*MB_Count-(MB_Summary+1)) div MB_Count)-1
    else
      Bar_Length:=((MB_Width-(MB_Item+2)-(MB_Value+1)*MB_Count) div MB_Count)-1;

    top_bar:='+'+copy(hyphen_string,1,MB_Item)+'+';
    mid_bar:=top_bar;
    head_bar:=top_bar+copy(hyphen_string,1,display_width-length(top_bar)-1)+'+';

    if bar_length>0 then
      for mb_i:=1 to MB_Count do
      begin
        top_bar:=top_bar+copy(hyphen_String,1,MB_Value+1+Bar_Length)+'+';
        mid_bar:=mid_bar+copy(hyphen_String,1,MB_Value)+'+'+Copy(hyphen_string,1,Bar_Length)+'+';
      End
    else
      for mb_i:=1 to MB_Count do
      begin
        top_bar:=top_bar+copy(hyphen_String,1,MB_Value)+'+';
        mid_bar:=mid_bar+copy(hyphen_String,1,MB_Value)+'+';
      End;

    if MB_Summary>0 then
    Begin
      top_bar:=top_bar+Copy(hyphen_String,1,MB_Summary)+'+';
      mid_bar:=mid_bar+Copy(hyphen_String,1,MB_Summary)+'+';
    End;

    if bar_length>0 then
    Begin
      For mb_i:=0 to MB_Count-1 do
        Titles[mb_i]:=Titles[mb_i]+copy(space_string,1,bar_length+2+MB_Value-Length(titles[mb_i]));
      Header:=Header+Copy(space_string,1,mb_count*(bar_length+2+mb_value)-length(header));
    End;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure remove_bits_detailed_output(var ToOutput : Text);
  //============================================================================
  var
    rb_i, rb_j, rb_k, rb_l : Integer;
    ss_k : Integer;
    btr_row_tot : Double;
    channel : Integer;
    time_string : String;
  //============================================================================
  Begin
  //============================================================================
    Write(tooutput,CRLF+
                   'Detailed bits-to-remove data per channel per codec-block.'+CRLF);

    ss_k:=(Display_Width-(8+2)-(3+1)) div (3*Global.Channels);

    titles[0]:='';
    titles[1]:='';
    titles[2]:='';
    header:='';
    Make_Bars(Display_Width,8,(3*Global.Channels-1),ss_k,3);

    write(ToOutput,Top_Bar,CRLF,'|  Time  |');
    for rb_i:=0 to ss_k-1 do
      for channel:=0 to Global.channels-1 do
      begin
        write(ToOutput,'#',channel);
        if channel=Global.channels-1 then
          write(ToOutput,'|')
        else
          write(ToOutput,',');
      end;
    write(ToOutput,'Tot|',CRLF,Mid_Bar,CRLF);

    ss_k:=ss_k*Global.Channels;

    btr_row_tot:=0;
    rb_j:=0;
    for rb_i:=0 to Global.blocks_processed-1 do
      for channel:=0 to Global.channels-1 do
      begin
        rb_l:=rb_j mod ss_k;

        if rb_l=0 then
        begin
          time_string_make(time_string,rb_i*Global.codec_block_size/Global.sample_rate);
          btr_row_tot:=0;
          write(ToOutput,'|',time_string,'|');
        end;

        rb_k:=bit_removal_history^[rb_j];
        btr_row_tot:=btr_row_tot+rb_k;
        write(ToOutput,rb_k:2);

        if channel=Global.channels-1 then
          write(ToOutput,'|')
        else
          write(ToOutput,',');

        if rb_l=ss_k-1 then
        begin
          write(ToOutput,btr_row_tot:3:0,'|',CRLF);
        end;
        inc(rb_j);
      end;

      rb_l:=rb_j mod ss_k;
      if (rb_l<ss_k-1) and (rb_l>0) then
      begin
        for rb_i:=0 to (ss_k-rb_l-1) div Global.channels do
          for channel:=0 to Global.channels-1 do
            if channel<Global.channels-1 then
              write(ToOutput,'   ')
            else
              write(ToOutput,'  |');
        write(ToOutput,btr_row_tot:3:0,'|',CRLF);
      end;
    Write(ToOutput,Mid_Bar,CRLF);
  //============================================================================
  End;
  //============================================================================

  {$IFNDEF USEASM}
  Function Bit_Scan_Left(BSL_Int : Integer) : Integer;
  Asm
                bsf     bsl_int,eax
  End;

  Function Bit_Scan_Right(BSR_Int : Integer) : Integer;
  Asm
                bsr     bsr_int,eax
  End;

  Procedure lsb_analysis;
  Var
    sa_i, sa_j  : Integer;
    temp_val    : Integer;
  Begin
    For sa_j:=0 to Global.Channels-1 do
    Begin
      DATA_block_value:=0;
      for sa_i:=0 to BlockData.Size.This-1 do
      Begin
        Temp_Val:=abs(BlockData.Pointer[2]^[sa_j,sa_i]);
        if temp_val=0 then
        Begin
          inc(DATA_sample_lsb[sa_j,0]);
          inc(DATA_sample_msb[sa_j,0]);
        End
        else
        begin
          inc(DATA_sample_lsb[sa_j,Bit_Scan_Left(temp_val)+1]);
          inc(DATA_sample_msb[sa_j,Bit_Scan_Right(temp_val)+1]);
        end;
        DATA_block_value:=DATA_block_value or Temp_Val;
      End;
      if DATA_block_value=0 then
      Begin
        inc(DATA_block_lsb[sa_j,0]);
        inc(DATA_block_msb[sa_j,0]);
      End
      else
      begin
        inc(DATA_block_lsb[sa_j,Bit_Scan_Left(DATA_block_value)+1]);
        inc(DATA_block_msb[sa_j,Bit_Scan_Right(DATA_block_value)+1]);
      end;

      BTRD_block_value:=0;
      for sa_i:=0 to BlockData.Size.This-1 do
      Begin
        Temp_Val:=abs(BlockData.BTRD[sa_j,sa_i]);
        if temp_val=0 then
        Begin
          inc(BTRD_sample_lsb[sa_j,0]);
          inc(BTRD_sample_msb[sa_j,0]);
        End
        else
        begin
          inc(BTRD_sample_lsb[sa_j,Bit_Scan_Left(temp_val)+1]);
          inc(BTRD_sample_msb[sa_j,Bit_Scan_Right(temp_val)+1]);
        end;
        BTRD_block_value:=BTRD_block_value or Temp_Val;
      End;
      if BTRD_block_value=0 then
      Begin
        inc(BTRD_block_lsb[sa_j,0]);
        inc(BTRD_block_msb[sa_j,0]);
      End
      else
      begin
        inc(BTRD_block_lsb[sa_j,Bit_Scan_Left(BTRD_block_value)+1]);
        inc(BTRD_block_msb[sa_j,Bit_Scan_Right(BTRD_block_value)+1]);
      end;

      CORR_block_value:=0;
      for sa_i:=0 to BlockData.Size.This-1 do
      Begin
        Temp_Val:=abs(BlockData.CORR[sa_j,sa_i]);
        if temp_val=0 then
        Begin
          inc(CORR_sample_lsb[sa_j,0]);
          inc(CORR_sample_msb[sa_j,0]);
        End
        else
        begin
          inc(CORR_sample_lsb[sa_j,Bit_Scan_Left(temp_val)+1]);
          inc(CORR_sample_msb[sa_j,Bit_Scan_Right(temp_val)+1]);
        end;
        CORR_block_value:=CORR_block_value or Temp_Val;
      End;
      if CORR_block_value=0 then
      Begin
        inc(CORR_block_lsb[sa_j,0]);
        inc(CORR_block_msb[sa_j,0]);
      End
      else
      begin
        inc(CORR_block_lsb[sa_j,Bit_Scan_Left(CORR_block_value)+1]);
        inc(CORR_block_msb[sa_j,Bit_Scan_Right(CORR_block_value)+1]);
      end;
    End;
  End;
  {$ELSE}
  //============================================================================
  procedure lsb_analysis; assembler;
  //============================================================================
  asm
                push    eax
                push    ebp
                push    ebx
                push    ecx
                push    edi
                push    edx
                push    esi

                mov     DATA_block_value,0
                mov     BTRD_block_value,0

                mov     esi,Global.channels
                jmp     @@11

                //==============================================================
                // Channel Loop
                //==============================================================
      @@01:     mov     ebx,dword ptr BlockData.Channel_Offset[esi*POINTER_SIZE]
                mov     edx,dword ptr RESULTS_OFFSET[esi*POINTER_SIZE]
                mov     ecx,BlockData.Size.This
                mov     ebp,dword ptr BlockData.Pointer[2*POINTER_SIZE]
                add     ebp,ebx
                jmp     @@10

                  //============================================================
                  // Sample Loop
                  //============================================================
                  // Scan input sample
                  //============================================================
      @@02:       mov     eax,dword ptr [ebp+ecx*INTEGER_SIZE]
                  cmp     eax,0
                  jg      @@04
                  jl      @@03

                  inc     dword ptr DATA_sample_lsb[edx]
                  inc     dword ptr DATA_sample_msb[edx]
                  jmp     @@05

      @@03:       neg     eax

      @@04:       or      DATA_block_value,eax

                  bsf     edi,eax
                  add     dword ptr DATA_sample_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                  bsr     edi,eax
                  add     dword ptr DATA_sample_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

                  //============================================================
                  // Scan bit-removed sample
                  //============================================================
      @@05:       mov     eax,dword ptr BlockData.BTRD[ebx+ecx*INTEGER_SIZE]
                  cmp     eax,0
                  jg      @@07
                  jl      @@06

                  inc     dword ptr BTRD_sample_lsb[edx]
                  inc     dword ptr BTRD_sample_msb[edx]
                  jmp     @@08

      @@06:       neg     eax

      @@07:       or      BTRD_block_value,eax

                  bsf     edi,eax
                  add     dword ptr BTRD_sample_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                  bsr     edi,eax
                  add     dword ptr BTRD_sample_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

                  //============================================================
                  // Scan bit-removed sample
                  //============================================================
      @@08:       mov     eax,dword ptr BlockData.CORR[ebx+ecx*INTEGER_SIZE]
                  cmp     eax,0
                  jg      @@0A
                  jl      @@09

                  inc     dword ptr CORR_sample_lsb[edx]
                  inc     dword ptr CORR_sample_msb[edx]
                  jmp     @@10

      @@09:       neg     eax

      @@0A:       or      CORR_block_value,eax

                  bsf     edi,eax
                  add     dword ptr CORR_sample_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                  bsr     edi,eax
                  add     dword ptr CORR_sample_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

                  //============================================================
                  // End of Sample Loop
                  //============================================================
      @@10:       sub     ecx,1
                  jnb     @@02
                  //============================================================

                //==============================================================
                // End of Channel Loop
                //==============================================================
      @@11:     sub     esi,1
                jnb     @@01
                //==============================================================

                //==============================================================
                // Calculate lsb and msb per codec-block
                //==============================================================
                // Scan input sample
                //==============================================================
                mov     eax,DATA_block_value
                cmp     eax,0
                jg      @@21
                jl      @@20

                inc     dword ptr DATA_block_lsb[edx]
                inc     dword ptr DATA_block_msb[edx]
                jmp     @@22

      @@20:     neg     eax

      @@21:     bsf     edi,eax
                add     dword ptr DATA_block_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                bsr     edi,eax
                add     dword ptr DATA_block_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

                //==============================================================
                // Scan bit-removed sample
                //==============================================================
      @@22:     mov     eax,BTRD_block_value
                cmp     eax,0
                jg      @@24
                jl      @@23

                inc     dword ptr BTRD_block_lsb[edx]
                inc     dword ptr BTRD_block_msb[edx]
                jmp     @@25

      @@23:     neg     eax

      @@24:     bsf     edi,eax
                add     dword ptr BTRD_block_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                bsr     edi,eax
                add     dword ptr BTRD_block_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

                //==============================================================
                // Scan bit-removed sample
                //==============================================================
      @@25:     mov     eax,CORR_block_value
                cmp     eax,0
                jg      @@27
                jl      @@26

                inc     dword ptr CORR_block_lsb[edx]
                inc     dword ptr CORR_block_msb[edx]
                jmp     @@99

      @@26:     neg     eax

      @@27:     bsf     edi,eax
                add     dword ptr CORR_block_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                bsr     edi,eax
                add     dword ptr CORR_block_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

      @@99:     pop     esi
                pop     edx
                pop     edi
                pop     ecx
                pop     ebx
                pop     ebp
                pop     eax
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  //============================================================================
  Procedure Write_bitdist(var ToOutput : Text);
  //============================================================================
  var
    nd_i, nd_j : Integer;
    this_bit_removed_percentage : single;
    bits_removed_filled : Integer;
  //============================================================================
  begin
    if not Parameters.perchannel then
    Begin
      Header:='| Distribution of bits removed from each codec-block';
      Make_Bars(Display_Width,3,6,1);
      Write(ToOutput,CRLF+
                     'Bits removed distribution.',CRLF,
                     Head_Bar,CRLF,
                     '|Bit',Header,'|',CRLF,
                     mid_bar,CRLF);

      for nd_i:=0 to Global.Bits_Per_Sample-1 do
      begin
        this_bit_removed_percentage:=0;
        for nd_j:=0 to Global.Channels-1 do
          this_bit_removed_percentage:=this_bit_removed_percentage+stats.bits_removed[nd_j,nd_i]*Global.blocks_processed_recip*OneOver[Global.Channels];

        bits_removed_filled:=round(this_bit_removed_percentage*bar_length);

        Write(ToOutput,'|',nd_i:3);

        Write(ToOutput,'|',this_bit_removed_percentage*100:5:1,'%|');
        Write(ToOutput,copy(bits_filled,1,bits_removed_filled),copy(bits_empty,1,bar_length-bits_removed_filled),'|',CRLF);
      end;
      Write(ToOutput,mid_bar,CRLF);
    End
    else
    begin
      for nd_i:=0 to Global.Channels-1 do
        Titles[nd_i]:='| Ch. #'+char(48+nd_i);

      Make_Bars(Display_Width,3,6,Global.Channels);
      Write(ToOutput,CRLF+
                     'Distribution of bits removed from each codec-block / channel.',CRLF,
                     Top_Bar,CRLF,
                     '|Bit');
      for nd_i:=0 to Global.Channels-1 do
        Write(ToOutput,Titles[nd_i]);
      Write(ToOutput,'|',CRLF,Mid_Bar,CRLF);

      for nd_i:=0 to Global.Bits_Per_Sample-1 do
      begin
        Write(ToOutput,'|',nd_i:3,'|');
        for nd_j:=0 to Global.Channels-1 do
        begin
          this_bit_removed_percentage:=stats.bits_removed[nd_j,nd_i]*Global.blocks_processed_recip;

          bits_removed_filled:=round(this_bit_removed_percentage*bar_length);

          Write(ToOutput,this_bit_removed_percentage*100:5:1,'%|');

          Write(ToOutput,copy(bits_filled,1,bits_removed_filled),copy(bits_empty,1,bar_length-bits_removed_filled),'|');
        end;
        Write(toOutput,CRLF);
      end;
      write(ToOutput,Mid_Bar,CRLF);
    end;
  end;
  //============================================================================

  //============================================================================
  Procedure Write_blockdist(var ToOutput : Text);
  //============================================================================
  var
    nd_i, nd_j : Integer;
    bits_filled_wave                  : Integer;
    bits_filled_btrd                  : Integer;
    bits_filled_corr                  : Integer;
    this_DATA_percentage              : Double;
    this_BTRD_percentage              : Double;
    this_CORR_percentage              : Double;
  //============================================================================
  begin
    Titles[0]:='|Input lsb distribution';
    Titles[1]:='|Lossy lsb distribution';
    Titles[2]:='|LWCDF lsb distribution';
    Make_Bars(Display_Width,3,6,3);
    Write(ToOutput,CRLF+
                   'Codec-block least significant bit (lsb) distribution.',CRLF,
                   Top_Bar,CRLF,
                   '|Bit',Titles[0],Titles[1],Titles[2],'|',CRLF,
                   Mid_Bar,CRLF);

    for nd_i:=0 to Global.Bits_Per_Sample do
    begin
      this_DATA_percentage:=0;
      this_BTRD_percentage:=0;
      this_CORR_percentage:=0;
      for nd_j:=0 to Global.Channels-1 do
      Begin
        this_DATA_percentage:=this_DATA_percentage+DATA_block_lsb[nd_j,nd_i]*Global.blocks_processed_recip;
        this_BTRD_percentage:=this_BTRD_percentage+BTRD_block_lsb[nd_j,nd_i]*Global.blocks_processed_recip;
        this_CORR_percentage:=this_CORR_percentage+CORR_block_lsb[nd_j,nd_i]*Global.blocks_processed_recip;
      End;
      bits_filled_wave:=round(this_DATA_percentage*bar_length);
      bits_filled_btrd:=round(this_BTRD_percentage*bar_length);
      bits_filled_CORR:=round(this_CORR_percentage*bar_length);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_DATA_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave),copy(bits_empty,1,bar_length-bits_filled_wave),'|');
      Write(ToOutput,floattostrf(this_BTRD_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd),copy(bits_empty,1,bar_length-bits_filled_btrd),'|');
      Write(ToOutput,floattostrf(this_CORR_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_CORR),copy(bits_empty,1,bar_length-bits_filled_CORR),'|',CRLF);
    end;

    Titles[0]:='|Input msb distribution';
    Titles[1]:='|Lossy msb distribution';
    Titles[2]:='|LWCDF msb distribution';
    Make_Bars(Display_Width,3,6,3);
    Write(ToOutput,Mid_Bar,CRLF,CRLF,
                   'Codec-block most significant bit (msb) distribution.',CRLF,
                   Top_Bar,CRLF,
                   '|Bit',Titles[0],Titles[1],Titles[2],'|',CRLF,
                   Mid_Bar,CRLF);

    for nd_i:=0 to Global.Bits_Per_Sample do
    begin
      this_DATA_percentage:=0;
      this_BTRD_percentage:=0;
      this_CORR_percentage:=0;
      for nd_j:=0 to Global.Channels-1 do
      Begin
        this_DATA_percentage:=this_DATA_percentage+DATA_block_msb[nd_j,nd_i]*Global.blocks_processed_recip;
        this_BTRD_percentage:=this_BTRD_percentage+BTRD_block_msb[nd_j,nd_i]*Global.blocks_processed_recip;
        this_CORR_percentage:=this_CORR_percentage+CORR_block_msb[nd_j,nd_i]*Global.blocks_processed_recip;
      End;

      bits_filled_wave:=round(this_DATA_percentage*bar_length);
      bits_filled_btrd:=round(this_BTRD_percentage*bar_length);
      bits_filled_CORR:=round(this_CORR_percentage*bar_length);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_DATA_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave),copy(bits_empty,1,bar_length-bits_filled_wave),'|');
      Write(ToOutput,floattostrf(this_BTRD_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd),copy(bits_empty,1,bar_length-bits_filled_btrd),'|');
      Write(ToOutput,floattostrf(this_CORR_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_CORR),copy(bits_empty,1,bar_length-bits_filled_CORR),'|',CRLF);
    end;

    Write(ToOutput,mid_bar,CRLF);

  end;
  //============================================================================

  //============================================================================
  Procedure Write_SampleDist(var ToOutput : Text);
  //============================================================================
  var
    nd_i, nd_j : Integer;
    bits_filled_wave                  : Integer;
    bits_filled_btrd                  : Integer;
    bits_filled_CORR                  : Integer;
    this_DATA_percentage              : Double;
    this_BTRD_percentage              : Double;
    this_CORR_percentage              : Double;
  //============================================================================
  begin
    Titles[0]:='|Input lsb distribution';
    Titles[1]:='|Lossy lsb distribution';
    Titles[2]:='|LWCDF lsb distribution';
    Make_Bars(Display_Width,3,6,3);
    Write(ToOutput,CRLF+
                   'Sample least significant bit (lsb) distribution.'+CRLF+
                   Top_Bar,CRLF,
                   '|Bit',Titles[0],Titles[1],Titles[2],'|',CRLF,
                   Mid_Bar,CRLF);
    for nd_i:=0 to Global.Bits_Per_Sample do
    begin
      this_DATA_percentage:=0;
      this_BTRD_percentage:=0;
      this_CORR_percentage:=0;

      for nd_j:=0 to Global.Channels-1 do
      Begin
        this_DATA_percentage:=this_DATA_percentage+DATA_sample_lsb[nd_j,nd_i]*Global.Samples_Processed_Recip;
        this_BTRD_percentage:=this_BTRD_percentage+BTRD_sample_lsb[nd_j,nd_i]*Global.Samples_Processed_Recip;
        this_CORR_percentage:=this_CORR_percentage+CORR_sample_lsb[nd_j,nd_i]*Global.Samples_Processed_Recip;
      End;

      bits_filled_wave:=round(this_DATA_percentage*bar_length);
      bits_filled_btrd:=round(this_BTRD_percentage*bar_length);
      bits_filled_CORR:=round(this_CORR_percentage*bar_length);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_DATA_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave),copy(bits_empty,1,bar_length-bits_filled_wave),'|');
      Write(ToOutput,floattostrf(this_BTRD_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd),copy(bits_empty,1,bar_length-bits_filled_btrd),'|');
      Write(ToOutput,floattostrf(this_CORR_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_CORR),copy(bits_empty,1,bar_length-bits_filled_CORR),'|',CRLF);
    end;

    Titles[0]:='|Input msb distribution';
    Titles[1]:='|Lossy msb distribution';
    Titles[2]:='|LWCDF msb distribution';
    Make_Bars(Display_Width,3,6,3);
    Write(ToOutput,Mid_Bar,CRLF,CRLF,
                   'Sample most significant bit (msb) distribution.',CRLF,
                   Top_Bar,CRLF,
                   '|Bit',Titles[0],Titles[1],Titles[2],'|',CRLF,
                   Mid_Bar,CRLF);
    for nd_i:=0 to Global.Bits_Per_Sample do
    begin
      this_DATA_percentage:=0;
      this_BTRD_percentage:=0;
      this_CORR_percentage:=0;
      for nd_j:=0 to Global.Channels-1 do
      Begin
        this_DATA_percentage:=this_DATA_percentage+DATA_sample_msb[nd_j,nd_i]*Global.Samples_Processed_Recip;
        this_BTRD_percentage:=this_BTRD_percentage+BTRD_sample_msb[nd_j,nd_i]*Global.Samples_Processed_Recip;
        this_CORR_percentage:=this_CORR_percentage+CORR_sample_msb[nd_j,nd_i]*Global.Samples_Processed_Recip;
      End;

      bits_filled_wave:=round(this_DATA_percentage*bar_length);
      bits_filled_btrd:=round(this_BTRD_percentage*bar_length);
      bits_filled_CORR:=round(this_CORR_percentage*bar_length);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_DATA_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave),copy(bits_empty,1,bar_length-bits_filled_wave),'|');
      Write(ToOutput,floattostrf(this_BTRD_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd),copy(bits_empty,1,bar_length-bits_filled_btrd),'|');
      Write(ToOutput,floattostrf(this_CORR_percentage*100,fffixed,99,1):5,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_CORR),copy(bits_empty,1,bar_length-bits_filled_CORR),'|',CRLF);
    end;

    Write(ToOutput,Mid_Bar,CRLF);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure Write_Histogram(var ToOutput : Text);
  //============================================================================
  var
    nd_i          : Integer;
    total_samples : Integer;
    temp_val      : Integer;

    DATA_Max      : Integer;
    BTRD_Max      : Integer;
    CORR_Max      : Integer;

    DATA_val      : Double;
    BTRD_val      : Double;
    CORR_val      : Double;

    DATA_len      : Integer;
    BTRD_len      : Integer;
    CORR_len      : Integer;

    DATA_bar      : String;
    BTRD_bar      : String;
    CORR_bar      : String;
  //============================================================================
  begin
    Titles[0]:='|Input Value Histogram';
    Titles[1]:='|Lossy Value Histogram';
    Titles[2]:='|LWCDF Value Histogram';
    Make_Bars(Display_Width,3,7,3);

    Write(ToOutput,CRLF+
                   'Sample Value Histogram.'+CRLF+
                   Top_Bar,CRLF,
                   '|Bin',Titles[0],Titles[1],Titles[2],'|',CRLF,
                   Mid_Bar,CRLF);

    total_samples:=0;
    DATA_Max:=0;
    BTRD_Max:=0;
    CORR_Max:=0;

    for nd_i:=0 to History.Histogram_Length do
    begin
      temp_val:=History.Histogram_DATA[nd_i];
      total_samples:=total_samples+temp_val;
      if temp_val>DATA_Max then
        DATA_Max:=temp_val;

      temp_val:=History.Histogram_BTRD[nd_i];
      if temp_val>BTRD_Max then
        BTRD_Max:=temp_val;

      temp_val:=History.Histogram_CORR[nd_i];
      if temp_val>CORR_Max then
        CORR_Max:=temp_val;
    End;

    for nd_i:=0 to History.Histogram_Length do
    begin
      temp_val:=History.Histogram_DATA[nd_i];
      DATA_val:=temp_val/Total_Samples*100;
      DATA_len:=floor(nlog2(temp_val)/nlog2(DATA_Max)*bar_length);
      DATA_bar:=copy(bits_filled,1,DATA_len)+copy(bits_empty,1,bar_length-DATA_len);

      temp_val:=History.Histogram_BTRD[nd_i];
      BTRD_val:=temp_val/Total_Samples*100;
      BTRD_len:=floor(nlog2(temp_val)/nlog2(BTRD_Max)*bar_length);
      BTRD_bar:=copy(bits_filled,1,BTRD_len)+copy(bits_empty,1,bar_length-BTRD_len);

      temp_val:=History.Histogram_CORR[nd_i];
      CORR_val:=temp_val/Total_Samples*100;
      CORR_len:=floor(nlog2(temp_val)/nlog2(CORR_Max)*bar_length);
      CORR_bar:=copy(bits_filled,1,CORR_len)+copy(bits_empty,1,bar_length-CORR_len);

      Write(ToOutput,'|',nd_i-History.Histogram_Offset:3,'|',DATA_val:6:2,'%|',DATA_bar,'|',BTRD_val:6:2,'%|',BTRD_bar,'|',CORR_val:6:2,'%|',CORR_bar,'|',CRLF);
    end;

    Write(ToOutput,Mid_Bar,CRLF);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure Write_FreqDist(Var ToOutput : Text);
  //============================================================================
  var
    nt_i, nt_j      : integer;
    nt_x, nt_m, nt_r: Double;
    fr_l, fr_h, fr_m: Integer;

    inp_DATA_string : String;
    Out_BTRD_String : String;
    Out_CORR_String : String;

    nt_inp_DATA_val : double;
    nt_Out_BTRD_val : double;
    nt_Out_CORR_val : double;

    nt_inp_DATA_len : integer;
    nt_Out_BTRD_len : integer;
    nt_Out_CORR_len : integer;

    BinChar         : Char;
  //============================================================================
  Begin
  //============================================================================
    if (not parameters.postanalyse) and (Global.Channels=2) then
      nt_x:=Global.Blocks_Processed_Recip
    else
      nt_x:=oneover[Global.Channels]*Global.blocks_processed_recip;

    nt_m:=-db_lg2x20*(Global.Bits_Per_Sample+log2(1.5)*SINGLE_HALF);
    nt_r:=1/nt_m;

    fr_l:=max(1,round(Global.Lower_Freq_Limit/Global.Sample_Rate*History.FFT.Length));
    fr_h:=round(Global.Upper_Freq_Limit/Global.Sample_Rate*History.FFT.Length);
    fr_m:=round(20000/Global.Sample_Rate*History.FFT.Length);

    write(ToOutput,CRLF+'Frequency Analysis of audio data.',CRLF);

    if Parameters.postanalyse then
    begin
      Titles[0]:='| Input Average (dBFS)';
      Titles[1]:='| Lossy Average (dBFS)';
      Titles[2]:='| LWCDF Average (dBFS)';
      Make_Bars(Display_Width,4,6,3,6);


      write(ToOutput,Top_Bar,CRLF,'| Bin',Titles[0],Titles[1],Titles[2],'|Change|',CRLF,mid_bar,CRLF);

      for nt_i:=0 to History.FFT.Length_Half do
      begin
        nt_inp_DATA_val:=0;
        nt_Out_BTRD_val:=0;
        nt_Out_CORR_val:=0;

        for nt_j:=0 to Global.Channels-1 do
        Begin
          nt_inp_DATA_val:=nt_inp_DATA_val+history.DATA_results[nt_j,nt_i];
          nt_Out_BTRD_val:=nt_Out_BTRD_val+history.BTRD_results[nt_j,nt_i];
          nt_Out_CORR_val:=nt_Out_CORR_val+history.CORR_results[nt_j,nt_i];
        End;

        nt_inp_DATA_val:=max(nt_m,db_lg2x20*(nlog2(nt_inp_DATA_val*nt_x)*SINGLE_HALF-History.FFT.Bit_Length+3-Global.Bits_Per_Sample));
        nt_Out_BTRD_val:=max(nt_m,db_lg2x20*(nlog2(nt_Out_BTRD_val*nt_x)*SINGLE_HALF-History.FFT.Bit_Length+3-Global.Bits_Per_Sample));
        nt_Out_CORR_val:=max(nt_m,db_lg2x20*(nlog2(nt_Out_CORR_val*nt_x)*SINGLE_HALF-History.FFT.Bit_Length+3-Global.Bits_Per_Sample));

        nt_inp_DATA_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_inp_DATA_val*nt_r)));
        nt_Out_BTRD_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_Out_BTRD_val*nt_r)));
        nt_Out_CORR_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_Out_CORR_val*nt_r)));

        inp_DATA_string:=copy(bits_filled,1,nt_inp_DATA_len)+copy(bits_empty,1,bar_length-nt_inp_DATA_len);
        Out_BTRD_string:=copy(bits_filled,1,nt_Out_BTRD_len)+copy(bits_empty,1,bar_length-nt_Out_BTRD_len);
        Out_CORR_string:=copy(bits_filled,1,nt_Out_CORR_len)+copy(bits_empty,1,bar_length-nt_Out_CORR_len);

        if nt_i=fr_l then
          BinChar:='L'
        else
          if nt_i=fr_h then
            BinChar:='U'
          else
            if nt_i=fr_m then
              BinChar:='*'
            else
              BinChar:=' ';

        if nt_i=0 then
          Write(ToOutput,'| DC ')
        else
          if nt_i=powersof.twoint[History.FFT.bit_length-1] then
            write(ToOutput,'|FS/2')
          else
            write(ToOutput,'|',BinChar,nt_i:3);

        write(Tooutput,'|',nt_inp_DATA_val:6:1,'|',inp_DATA_string,'|',
                           nt_Out_BTRD_val:6:1,'|',Out_BTRD_string,'|',
                           nt_Out_CORR_val:6:1,'|',Out_CORR_string,'|',
                           BinChar,nRound10(nt_Out_BTRD_val-nt_inp_DATA_val,2):5:2,'|',CRLF);
      End;
      write(ToOutput,Mid_Bar,CRLF)
    end
    else
      if global.channels<>2 then
      begin
        Titles[0]:='| Input Average (dBFS)';
        Make_Bars(Display_Width,4,6,1);

        write(ToOutput,Top_Bar,CRLF,'| Bin',Titles[0],'|'+CRLF+Mid_Bar,CRLF);

        for nt_i:=0 to History.FFT.Length_Half do
        begin
          nt_inp_DATA_val:=0;

          for nt_j:=0 to Global.Channels-1 do
            nt_inp_DATA_val:=nt_inp_DATA_val+history.DATA_results[nt_j,nt_i];

          nt_inp_DATA_val:=max(nt_m,db_lg2x20*(nlog2(nt_inp_DATA_val*nt_x)*SINGLE_HALF-History.FFT.Bit_Length+3-Global.Bits_Per_Sample));

          nt_inp_DATA_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_inp_DATA_val*nt_r)));

          inp_DATA_string:=copy(bits_filled,1,nt_inp_DATA_len)+copy(bits_empty,1,bar_length-nt_inp_DATA_len);

          if nt_i=fr_l then
            BinChar:='L'
          else
            if nt_i=fr_h then
              BinChar:='U'
            else
              if nt_i=fr_m then
                BinChar:='*'
              else
                BinChar:=' ';

          if nt_i=0 then
            Write(ToOutput,'| DC ')
          else
            if nt_i=powersof.twoint[History.FFT.bit_length-1] then
              write(ToOutput,'|FS/2')
            else
              write(ToOutput,'|',BinChar,nt_i:3);

          write(ToOutput,'|',nt_inp_DATA_val:6:1,'|',inp_DATA_string,'|',CRLF);
        end;
        write(ToOutput,mid_bar,CRLF);
      end
    Else
      begin
        Titles[0]:='| Input Average (dBFS) Channel #0';
        Titles[1]:='| Input Average (dBFS) Channel #1';
        Make_Bars(Display_Width,4,6,2);

        write(ToOutput,Top_Bar,CRLF,'| Bin',Titles[0],Titles[1],'|'+CRLF+Mid_Bar,CRLF);

        for nt_i:=0 to History.FFT.Length_Half do
        begin
          nt_inp_DATA_val:=history.DATA_results[0,nt_i];
          nt_out_BTRD_val:=history.DATA_results[1,nt_i];

          nt_inp_DATA_val:=max(nt_m,db_lg2x20*(nlog2(nt_inp_DATA_val*nt_x)*SINGLE_HALF-History.FFT.Bit_Length+3-Global.Bits_Per_Sample));
          nt_out_BTRD_val:=max(nt_m,db_lg2x20*(nlog2(nt_out_BTRD_val*nt_x)*SINGLE_HALF-History.FFT.Bit_Length+3-Global.Bits_Per_Sample));

          nt_inp_DATA_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_inp_DATA_val*nt_r)));
          nt_out_BTRD_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_out_BTRD_val*nt_r)));

          inp_DATA_string:=copy(bits_filled,1,nt_inp_DATA_len)+copy(bits_empty,1,bar_length-nt_inp_DATA_len);
          out_BTRD_string:=copy(bits_filled,1,nt_out_BTRD_len)+copy(bits_empty,1,bar_length-nt_out_BTRD_len);

          if nt_i=fr_l then
            BinChar:='L'
          else
            if nt_i=fr_h then
              BinChar:='U'
            else
              if nt_i=fr_m then
                BinChar:='*'
              else
                BinChar:=' ';

          if nt_i=0 then
            Write(ToOutput,'| DC ')
          else
            if nt_i=powersof.twoint[History.FFT.bit_length-1] then
              write(ToOutput,'|FS/2')
            else
              write(ToOutput,'|',BinChar,nt_i:3);

          write(ToOutput,'|',nt_inp_DATA_val:6:1,'|',inp_DATA_string,'|',nt_out_BTRD_val:6:1,'|',out_BTRD_string,'|',CRLF);
        end;
        write(ToOutput,mid_bar,CRLF);
      end;
    Write(ToOutput,'Legend: L = Lower Frequency Calculation Limit ( '+floattostrf(Global.Lower_Freq_Limit*OneOver[1000],fffixed,99,3)+'kHz);',CRLF,
                   '        U = Upper Frequency Calculation Limit ('+floattostrf(Global.Upper_Freq_Limit*OneOver[1000],fffixed,99,3)+'kHz);',CRLF,
                   '        * = 20.000kHz'+CRLF);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Write_LongDist(Var ToOutput : Text);
  //============================================================================
  var
    nt_i, nt_j      : integer;
    nt_x, nt_m, nt_r: Double;
    fr_l, fr_h, fr_m: Integer;

    inp_DATA_string : String;
    Out_BTRD_String : String;
    Out_CORR_String : String;

    nt_inp_DATA_val : double;
    nt_Out_BTRD_val : double;
    nt_Out_CORR_val : double;

    nt_inp_DATA_len : integer;
    nt_Out_BTRD_len : integer;
    nt_Out_CORR_len : integer;

    BinChar         : Char;
  //============================================================================
  Begin
  //============================================================================
    nt_x:=oneover[Global.Channels]*Global.blocks_processed_recip;

    nt_m:=-db_lg2x20*(Global.Bits_Per_Sample+log2(1.5)*SINGLE_HALF);
    nt_r:=1/nt_m;

    fr_l:=max(1,round(Global.Lower_Freq_Limit/Global.Sample_Rate*LongDist.FFT.Length));
    fr_h:=round(Global.Upper_Freq_Limit/Global.Sample_Rate*LongDist.FFT.Length);
    fr_m:=round(20000/Global.Sample_Rate*LongDist.FFT.Length);

    write(ToOutput,CRLF+'Frequency Analysis of audio data.',CRLF);

    Titles[0]:='| Input Average (dBFS)';
    Titles[1]:='| Lossy Average (dBFS)';
    Titles[2]:='| LWCDF Average (dBFS)';
    Make_Bars(Display_Width,4,6,3,6);


    write(ToOutput,Top_Bar,CRLF,'| Bin',Titles[0],Titles[1],Titles[2],'|Change|',CRLF,mid_bar,CRLF);

    for nt_i:=0 to LongDist.FFT.Length_Half do
    begin
      nt_inp_DATA_val:=0;
      nt_Out_BTRD_val:=0;
      nt_Out_CORR_val:=0;

      for nt_j:=0 to Global.Channels-1 do
      Begin
        nt_inp_DATA_val:=nt_inp_DATA_val+LongDist.DATA_results[nt_j,nt_i];
        nt_Out_BTRD_val:=nt_Out_BTRD_val+LongDist.BTRD_results[nt_j,nt_i];
        nt_Out_CORR_val:=nt_Out_CORR_val+LongDist.CORR_results[nt_j,nt_i];
      End;

      nt_inp_DATA_val:=max(nt_m,db_lg2x20*(nlog2(nt_inp_DATA_val*nt_x)*SINGLE_HALF-LongDist.FFT.Bit_Length+3-Global.Bits_Per_Sample));
      nt_Out_BTRD_val:=max(nt_m,db_lg2x20*(nlog2(nt_Out_BTRD_val*nt_x)*SINGLE_HALF-LongDist.FFT.Bit_Length+3-Global.Bits_Per_Sample));
      nt_Out_CORR_val:=max(nt_m,db_lg2x20*(nlog2(nt_Out_CORR_val*nt_x)*SINGLE_HALF-LongDist.FFT.Bit_Length+3-Global.Bits_Per_Sample));

      nt_inp_DATA_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_inp_DATA_val*nt_r)));
      nt_Out_BTRD_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_Out_BTRD_val*nt_r)));
      nt_Out_CORR_len:=bar_length-min(bar_length,max(0,round(bar_length*nt_Out_CORR_val*nt_r)));

      inp_DATA_string:=copy(bits_filled,1,nt_inp_DATA_len)+copy(bits_empty,1,bar_length-nt_inp_DATA_len);
      Out_BTRD_string:=copy(bits_filled,1,nt_Out_BTRD_len)+copy(bits_empty,1,bar_length-nt_Out_BTRD_len);
      Out_CORR_string:=copy(bits_filled,1,nt_Out_CORR_len)+copy(bits_empty,1,bar_length-nt_Out_CORR_len);

      if nt_i=fr_l then
        BinChar:='L'
      else
        if nt_i=fr_h then
          BinChar:='U'
        else
          if nt_i=fr_m then
            BinChar:='*'
          else
            BinChar:=' ';

      if nt_i=0 then
        Write(ToOutput,'| DC ')
      else
        if nt_i=powersof.twoint[LongDist.FFT.bit_length-1] then
          write(ToOutput,'|FS/2')
        else
          write(ToOutput,'|',BinChar,nt_i:3);

      write(Tooutput,'|',nt_inp_DATA_val:6:1,'|',inp_DATA_string,'|',
                         nt_Out_BTRD_val:6:1,'|',Out_BTRD_string,'|',
                         nt_Out_CORR_val:6:1,'|',Out_CORR_string,'|',
                         BinChar,nRound10(nt_Out_BTRD_val-nt_inp_DATA_val,2):5:2,'|',CRLF);
    End;
    write(ToOutput,Mid_Bar,CRLF);

    Write(ToOutput,'Legend: L = Lower Frequency Calculation Limit ( '+floattostrf(Global.Lower_Freq_Limit*OneOver[1000],fffixed,99,3)+'kHz);',CRLF,
                   '        U = Upper Frequency Calculation Limit ('+floattostrf(Global.Upper_Freq_Limit*OneOver[1000],fffixed,99,3)+'kHz);',CRLF,
                   '        * = 20.000kHz'+CRLF);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Write_Min_Bin_Dist(Var ToOutput : Text);
  //============================================================================
  Var
    nt_i, nt_j      : Integer;
    Bar_Full_Len    : Integer;

    Old_Percent     : Double;
    New_Percent     : Double;
    Alt_Percent     : Double;

    Tot_Old_Percent : Double;
    Tot_New_Percent : Double;
    Tot_Alt_Percent : Double;

    Old_Bar_Str     : String;
    New_Bar_Str     : String;
    Alt_Bar_Str     : String;

    Local_Max       : integer;
    Percent_Used    : Array[1..PRECALC_ANALYSES+2] of Double;
    fr_l, fr_h, fr_m: Integer;
    BinChar         : Char;
  Begin
  //============================================================================
    If Parameters.Spread=2 then
    Begin
      Titles[0]:='| Old Minimum Value Used';
      Titles[1]:='| New Minimum Value Used';
      Titles[2]:='| Minimum Value Used';

      Make_Bars(Display_Width,4,7,3);

      for nt_j:=1 to precalc_analyses do
      Begin
        if (Settings.FFT_Analysis_Switches and Global.Analysis.Length[nt_j])>0 then
      End;

      for nt_j:=1 to precalc_analyses do
        if (Settings.FFT_Analysis_Switches and Global.Analysis.Length[nt_j])>0 then
        Begin
          local_max:=0;
          tot_old_percent:=0;
          tot_new_percent:=0;
          tot_alt_percent:=0;

          for nt_i:=0 to Spreading.Bins.Hi[Global.Analysis.Bits[nt_j]]+1 do
            local_max:=max(local_max,Process.Old_Min_Used_History[nt_j,nt_i]+Process.New_Min_Used_History[nt_j,nt_i]);

          fr_l:=Spreading.Bins.Lo[Global.Analysis.Bits[nt_j]];
          fr_h:=Spreading.Bins.Hi[Global.Analysis.Bits[nt_j]];
          fr_m:=round(20000/Global.Sample_Rate*Global.Analysis.Length[nt_j]);

          Write(ToOutput,CRLF+'Spreading Algorithm Results : Minima, FFT Length = '+inttostr(Global.Analysis.Length[nt_j]) +CRLF+
                         Top_Bar+CRLF+
                         '|Bin ',Titles[0],Titles[1],'|'+CRLF+
                         Mid_Bar+CRLF);
          for nt_i:=0 to Spreading.Bins.Hi[Global.Analysis.Bits[nt_j]]+1 do
          Begin
            if nt_i=fr_l then
              BinChar:='L'
            else
              if nt_i=fr_h then
                BinChar:='U'
              else
                if nt_i=fr_m then
                  BinChar:='*'
                else
                  BinChar:=' ';

            if nt_i=0 then
              Write(ToOutput,'| DC ')
            else
              if nt_i=Global.Analysis.Length[nt_j] then
                write(ToOutput,'|FS/2')
              else
                write(ToOutput,'|',BinChar,nt_i:3);

            Old_Percent:=Process.Old_Min_Used_History[nt_j,nt_i]/Process.Analyses_Completed[nt_j]*100;
            New_Percent:=Process.New_Min_Used_History[nt_j,nt_i]/Process.Analyses_Completed[nt_j]*100;
            Alt_Percent:=Old_Percent+New_Percent;

            Bar_Full_Len:=round(Process.Old_Min_Used_History[nt_j,nt_i]/Local_Max*bar_length);
            Old_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

            Bar_Full_Len:=round(Process.New_Min_Used_History[nt_j,nt_i]/Local_Max*bar_length);
            New_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

            Bar_Full_Len:=round((Process.Old_Min_Used_History[nt_j,nt_i]+Process.New_Min_Used_History[nt_j,nt_i])/Local_Max*bar_length);
            Alt_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

            Tot_Old_Percent:=Tot_Old_Percent+Old_Percent;
            Tot_New_Percent:=Tot_New_Percent+New_Percent;
            Tot_Alt_Percent:=Tot_Alt_Percent+Alt_Percent;

            write(ToOutput,'|',Old_Percent:6:2,'%|',Old_Bar_Str,'|',new_Percent:6:2,'%|',New_Bar_Str,'|',alt_percent:6:2,'%|',Alt_Bar_Str,'|',CRLF);
          End;

          Write(ToOutput,Mid_Bar,CRLF);

          Bar_Full_Len:=round(Tot_Old_Percent*bar_length*OneOver[100]);
          Old_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

          Bar_Full_Len:=round(Tot_New_Percent*bar_length*OneOver[100]);
          New_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

          Bar_Full_Len:=round(Tot_Alt_Percent*bar_length*OneOver[100]);
          Alt_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

          write(ToOutput,'|Tot.|',Tot_Old_Percent:6:2,'%|',Old_Bar_Str,'|',Tot_new_Percent:6:2,'%|',New_Bar_Str,'|',Tot_Alt_Percent:6:2,'%|',Alt_Bar_Str,'|',CRLF);

          Write(ToOutput,Mid_Bar,CRLF);
        End;
    End;
    //==========================================================================
    Titles[0]:='| Minimum used per codec-block';
    Make_Bars(Display_Width,4,7,1);

    Write(ToOutput,CRLF+'Minimum used to determine bits-to-remove, by FFT length:'+CRLF+
                   Top_Bar,CRLF,
                   '|FFT ',Titles[0],'|'+CRLF+
                   Mid_Bar+CRLF);

    for nt_j:=1 to PRECALC_ANALYSES do
      if (Settings.FFT_Analysis_Switches and Global.Analysis.Length[nt_j])>0 then
      Begin
        Percent_Used[nt_j]:=0;
        for nt_i:=0 to global.channels-1 do
          Percent_Used[nt_j]:=Percent_Used[nt_j]+results.minima[nt_i,nt_j];
        Percent_Used[nt_j]:=Percent_Used[nt_j]*Global.Blocks_Processed_Recip*OneOver[Global.Channels]*100;
        Bar_Full_Len:=round(Percent_Used[nt_j]*bar_length*oneover[100]);
        old_bar_str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);
        Write(ToOutput,'|',Global.Analysis.Length[nt_j]:4,'|',Percent_Used[nt_j]:6:2,'%|',Old_Bar_Str,'|',CRLF);
      End;

    nt_j:=7;
    Percent_Used[nt_j]:=0;
    for nt_i:=0 to global.channels-1 do
      Percent_Used[nt_j]:=Percent_Used[nt_j]+results.minima[nt_i,nt_j];
    Percent_Used[nt_j]:=Percent_Used[nt_j]*Global.Blocks_Processed_Recip*OneOver[Global.Channels]*100;
    Bar_Full_Len:=round(Percent_Used[nt_j]*bar_length*oneover[100]);
    old_bar_str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);
    Write(ToOutput,'|Stat|',Percent_Used[nt_j]:6:2,'%|',Old_Bar_Str,'|',CRLF);

    nt_j:=8;
    Percent_Used[nt_j]:=0;
    for nt_i:=0 to global.channels-1 do
      Percent_Used[nt_j]:=Percent_Used[nt_j]+results.minima[nt_i,nt_j];
    Percent_Used[nt_j]:=Percent_Used[nt_j]*Global.Blocks_Processed_Recip*OneOver[Global.Channels]*100;
    Bar_Full_Len:=round(Percent_Used[nt_j]*bar_length*oneover[100]);
    old_bar_str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);
    Write(ToOutput,'|Dyn.|',Percent_Used[nt_j]:6:2,'%|',Old_Bar_Str,'|',CRLF);

    Write(ToOutput,Mid_Bar,CRLF);

    //==========================================================================
    Titles[0]:='| Old Minimum Value Used';
    Titles[1]:='| New Minimum Value Used';
    Titles[2]:='| Average Value Used';
    Make_Bars(Display_Width,4,7,3);

    Write(ToOutput,CRLF+'Spreading Algorithm Results : Minima, per analysis'+CRLF+
                   Top_Bar+CRLF+
                   '|FFT ',Titles[0],Titles[1],Titles[2],'|'+CRLF+
                   Mid_Bar+CRLF);

    For nt_j:=1 to precalc_analyses do
      if (Settings.FFT_Analysis_Switches and Global.Analysis.Length[nt_j])>0 then
      Begin
        Old_Percent:=Process.Old_Min_Used[nt_j]/Process.Analyses_Completed[nt_j]*100;
        New_Percent:=Process.New_Min_Used[nt_j]/Process.Analyses_Completed[nt_j]*100;
        Alt_Percent:=Process.Alt_Ave_Used[nt_j]/Process.Analyses_Completed[nt_j]*100;

        Bar_Full_Len:=round(Old_Percent*bar_length*OneOver[100]);
        Old_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

        Bar_Full_Len:=round(New_Percent*bar_length*OneOver[100]);
        New_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

        Bar_Full_Len:=round(Alt_Percent*bar_length*OneOver[100]);
        Alt_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

        Write(ToOutput,'|',Global.Analysis.Length[nt_j]:4,'|',Old_Percent:6:2,'%|',Old_Bar_Str,'|',New_Percent:6:2,'%|',New_Bar_Str,'|',Alt_Percent:6:2,'%|',Alt_Bar_Str,'|'+CRLF);
      End;
    Write(ToOutput,Mid_Bar,CRLF);

    //==========================================================================
    Titles[0]:='| BTR > Static-Max-Bits-To-Remove';
    Titles[1]:='| BTR > Dynamic-Max-Bits-To-Remove';
    Make_Bars(Display_Width,4,7,2);

    Write(ToOutput,CRLF+'Minimum_Bits_To_Keep Results (BTR limited to relevant maximum)'+CRLF+
                   Top_Bar+CRLF+
                   '|FFT ',Titles[0],Titles[1],'|'+CRLF+
                   Mid_Bar+CRLF);

    For nt_j:=1 to precalc_analyses do
      if (Settings.FFT_Analysis_Switches and Global.Analysis.Length[nt_j])>0 then
      Begin
        Old_Percent:=Process.Over_Static[nt_j]/Process.Analyses_Completed[nt_j]*100;
        New_Percent:=Process.Over_Dynamic[nt_j]/Process.Analyses_Completed[nt_j]*100;

        Bar_Full_Len:=round(Old_Percent*bar_length*OneOver[100]);
        Old_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

        Bar_Full_Len:=round(New_Percent*bar_length*OneOver[100]);
        New_Bar_Str:=Copy(Bits_Filled,1,Bar_Full_Len)+Copy(Bits_Empty,1,Bar_Length-Bar_Full_Len);

        Write(ToOutput,'|',Global.Analysis.Length[nt_j]:4,'|',Old_Percent:6:2,'%|',Old_Bar_Str,'|',New_Percent:6:2,'%|',New_Bar_Str,'|'+CRLF);
      End;
    Write(ToOutput,Mid_Bar,CRLF);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Process_Output;
  //============================================================================
  Begin
    if (Parameters.verbosity) and (not Parameters.silent) and ((Global.last_print=0) or (Global.blocks_processed mod Global.Output_Blocks=0)) then
    begin
      GetTimer;
      time_string_make(strings.elapsed,timer.elapsed);

      Write(ErrOutput,#13,'Progress  :');

      if Global.Wave_Size=MAX_WAVE_SIZE then
      begin
        size_string_make(strings.size,1.0*Global.samples_processed*Global.channels*Global.Bytes_Per_Sample);
        Write(ErrOutput,' ',strings.size,';')
      end
      else
        Write(ErrOutput,Global.samples_processed/Global.Total_Samples*100:6:2,'%;');

      write(ErrOutput,OneOver[Global.channels]*Global.bits_removed*Global.blocks_processed_recip:7:4,' bits;',Global.processing_rate:6:2,'x;');

      Write(ErrOutput,' ',strings.elapsed);
      if (Global.samples_processed>0) then
      begin
        time_string_make(strings.estimate,min(86399,Global.Total_Samples/Global.samples_processed*timer.elapsed));
        Write(ErrOutput,'/',strings.estimate,' ');
      end;

      Global.last_print:=Global.blocks_processed;
    end;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Write_Results(Var ToOutput : Text);
  var
    bits_lost : Double;
  //============================================================================
  begin
    Write(ToOutput,'Results   :',
                    OneOver[Global.channels]*Global.bits_removed*Global.blocks_processed_recip:7:4,' bits; ',
                    Global.processing_rate:0:2,'x; ',
                    strings.elapsed);

    if FFTW_Initialised then
      write(ToOutput,'; [F]')
    else
      write(ToOutput,'; [I]');

    If Global.Jump_Location<>0 then
      Write(tooutput,'; J:',CardinaltoHex(Global.Jump_Location),';',Global.Jump_Location and 31);

    write(ToOutput,CRLF);

    if Global.bits_lost+Global.clipped_samples>0 then
    begin
      Write(ToOutput,'Clipping  :');
      bits_lost:=nround10(Global.bits_lost*OneOver[Global.channels]*Global.blocks_processed_recip,4);
      if bits_lost>0.00005 then
        Write(ToOutput,bits_lost:7:4,' bits lost,');
      Write(ToOutput,' ',Global.clipped_samples,' clip');
      if Global.clipped_samples<>1 then
        Write(ToOutput,'s');
      if bits_lost>0 then
        Write(ToOutput,' remaining.'+CRLF)
      else
        Write(ToOutput,'.'+CRLF);
    end;

    if (Parameters.adaptive_shaping<>-1) and (Skipped_Filters>0) then
    begin
      write(ToOutput,'Filters   : ',inttostr(skipped_filters),' adaptive noise shaping filter');
      if skipped_filters<>1 then
        write(ToOutput,'s');
      write(ToOutput,' skipped.',CRLF);
    end;

    Global.Samples_Processed_Recip:=OneOver[Global.Channels]/Global.Samples_Processed;

    if Parameters.BitDist then
      write_bitdist(ToOutput);

    if Parameters.blockdist then
      write_blockdist(ToOutput);

    if Parameters.sampledist then
      write_sampledist(ToOutput);

    if Parameters.freqdist then
      write_freqdist(ToOutput);

    if Parameters.longdist then
      write_longdist(ToOutput);

    if Parameters.detail then
      remove_bits_detailed_output(ToOutput);

    if Parameters.spread<>-1 then
      Write_Min_Bin_Dist(ToOutput);

    if Parameters.Histogram then
      Write_Histogram(ToOutput);

    Flush(ToOutput);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure write_cleanup;
  //============================================================================
  begin
    GetTimer();
    time_string_make(strings.elapsed,timer.elapsed);

    if (not Parameters.silent) then
    begin
      if (Parameters.verbosity) then
      begin
        Write(ErrOutput,#13+'                                                                               '+#13);
        Write_Results(ErrOutput);
      end
      else
      begin
        write(ErrOutput,OneOver[Global.channels]*Global.bits_removed*Global.blocks_processed_recip:7:4,';',Global.processing_rate:6:2,'x; ',strings.elapsed);

        If FFTW_Initialised then
          write(erroutput,'; [F]')
        else
          write(erroutput,'; [I]');

        if (Parameters.adaptive_shaping<>-1) and (skipped_filters>0) then
          write(erroutput,'; F',inttostr(Skipped_Filters));

        If Global.Jump_Location<>0 then
          Write(ErrOutput,'; J',Global.Jump_Location and 15);

        write(erroutput,CRLF);
      end;
    end;

    if (Parameters.writetolog) then
    begin
      time_string_make(strings.time,Global.samples_processed/Global.Sample_Rate);
      size_string_make(strings.size,1.0*Global.samples_processed*Global.channels*Global.Bytes_Per_Sample);

      open_log_file;

      Write(LogOutput,version_string,strings.version_short,lossyWAVhead1,
                      'Processed : ',Strings.DateStamp,CRLF,
                      'Settings  : ',Strings.Parameter,CRLF,
                      'Filename  : ',WavFilePrintName,CRLF,
                      'File Info : ',Global.Sample_Rate/1000:0:2,'kHz; ',Global.channels,' channel; ',Global.Bits_Per_Sample,' bit, ',strings.time,', ',strings.size,CRLF);
      Write_Results(LogOutput);
      Write(LogOutput,CRLF);

      close_log_file;
    end;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure nOutput_Init;
  //============================================================================
  Var
    nd_i, nd_j : Integer;
  //============================================================================
  Begin
  //============================================================================
    for nd_j:=0 to MAX_CHANNELS-1 do
      for nd_i:=0 to 33 do
      Begin
        DATA_block_lsb[nd_j,nd_i]:=0;
        DATA_block_msb[nd_j,nd_i]:=0;
        DATA_sample_lsb[nd_j,nd_i]:=0;
        DATA_sample_msb[nd_j,nd_i]:=0;

        BTRD_block_lsb[nd_j,nd_i]:=0;
        BTRD_block_msb[nd_j,nd_i]:=0;
        BTRD_sample_lsb[nd_j,nd_i]:=0;
        BTRD_sample_msb[nd_j,nd_i]:=0;

        CORR_block_lsb[nd_j,nd_i]:=0;
        CORR_block_msb[nd_j,nd_i]:=0;
        CORR_sample_lsb[nd_j,nd_i]:=0;
        CORR_sample_msb[nd_j,nd_i]:=0;
      End;
  //============================================================================
  End;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  nOutput_Init;

  if Parameters.detail then
    new(bit_removal_history);

  if Parameters.width=-1 then
    Display_Width:=79
  else
    Display_Width:=Parameters.width;
//==============================================================================

//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
  if bit_removal_history<>nil then
    dispose(bit_removal_history);
//==============================================================================

//==============================================================================
end.
//==============================================================================
