//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nParameter;
//==============================================================================
interface
//==============================================================================
Uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================
Const
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAVHead1=', Copyright (C) 2007-2011 Nick Currie. Copyleft.'+CRLF;
  lossyWAVhead2= 'This is free software under the GNU GPLv3+ license; There is NO WARRANTY, to'+CRLF+
                 'the extent permitted by law. <http://www.gnu.org/licenses/> for details.'+CRLF;
//==============================================================================
Function  WAVFilePrintName      : String;   // name of WAV file OR 'from STDIN' if STDINPUT=1.

procedure open_log_file;
procedure close_log_file;

//==============================================================================
implementation
//==============================================================================
uses
  nmaths,
  SysUtils,
  Windows,
  Math;
//==============================================================================
//nparameter constants, variables & types
//==============================================================================
const
  UNITNAME='nParameter';

                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_GPL =
                 'This program is free software: you can redistribute it and/or modify it under'+
            CRLF+'the terms of the GNU General Public License as published by the Free Software'+
            CRLF+'Foundation, either version 3 of the License, or (at your option) any later'+
            CRLF+'version.'+CRLF+
            CRLF+'This program is distributed in the hope that it will be useful,but WITHOUT ANY'+
            CRLF+'WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A'+
            CRLF+'PARTICULAR PURPOSE.  See the GNU General Public License for more details.'+CRLF+
            CRLF+'You should have received a copy of the GNU General Public License along with'+
            CRLF+'this program.  If not, see <http://www.gnu.org/licenses/>.'+CRLF;
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_process_description =
                 'Process Description:'+CRLF+
            CRLF+'lossyWAV is a near lossless audio processor which dynamically reduces the '+
            CRLF+'bitdepth of the signal on a block-by-block basis. Bitdepth reduction adds noise'+
            CRLF+'to the processed output. The amount of permissible added noise is based on'+
            CRLF+'analysis of the signal levels in the default frequency range 20Hz to 16kHz.'+CRLF+
            CRLF+'If signals above the upper limiting frequency are at an even lower level, they'+
            CRLF+'can be swamped by the added noise. This is usually inaudible, but the behaviour'+
            CRLF+'can be changed by specifying a different --limit (in the range 10kHz to 20kHz).'+CRLF+
            CRLF+'For many audio signals there is little content at very high frequencies and'+
            CRLF+'forcing lossyWAV to keep the added noise level lower than the content at these'+
            CRLF+'frequencies can increase the bitrate dramatically for no perceptible benefit.'+CRLF+
            CRLF+'The noise added by the process is shaped using an adaptive method provided by'+
            CRLF+'Sebastian Gesemann. This method, as implemented in lossyWAV, aims to use the'+
            CRLF+'signal itself as the basis of the filter used for noise shaping. Adaptive noise'+
            CRLF+'shaping is enabled by default.'+CRLF;
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_standard_help =
                 'Usage   : lossyWAV <input wav file> <options>'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Example : lossyWAV musicfile.wav'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Quality Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-q, --quality <t>    where t is one of the following (default = standard):'+
            CRLF+'    I, insane        highest quality output, suitable for transcoding;'+
            CRLF+'    E, extreme       higher quality output, suitable for transcoding;'+
            CRLF+'    H, high          high quality output, suitable for transcoding;'+
            CRLF+'    S, standard      default quality output, considered to be transparent;'+
            CRLF+'    C, economic      intermediate quality output, likely to be transparent;'+
            CRLF+'    P, portable      good quality output for DAP use, may not be transparent;'+
            CRLF+'    X, extraportable lowest quality output, not fully transparent.'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Standard Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-C, --correction     write correction file for processed WAV file; default=off.'+
            CRLF+'-f, --force          forcibly over-write output file if it exists; default=off.'+
            CRLF+'-h, --help           display help.'+
            CRLF+'-L, --longhelp       display extended help.'+
            CRLF+'-M, --merge          merge existing lossy.wav and lwcdf.wav files.'+
            CRLF+'-o, --outdir <t>     destination directory for the output file(s).'+
            CRLF+'-v, --version        display the lossyWAV version number.'+
            CRLF+'-w, --writetolog     create (or add to) lossyWAV.log in the output directory.'+CRLF;
  lossyWAV_advanced_help =
            CRLF+'Advanced Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-                    take WAV input from STDIN.'+
            CRLF+'-c, --check          check if WAV file has already been processed; default=off.'+
            CRLF+'                     errorlevel=16 if already processed, 0 if not.'+
            CRLF+'-q, --quality <n>    quality preset (-5.0<=n<=10.0); (-5=lowest, 10=highest;'+
            CRLF+'                     default=2.5; I=10; E=7.5; H=5; S=2.5; C=0; P=-2.5; X=-5).'+
            CRLF+'--, --stdout         write WAV output to STDOUT.'+
            CRLF+'    --stdinname <t>  pseudo filename to use when input from STDIN.'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Advanced Quality Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-A, --adaptive <n/t> modify settings for Sebastian Gesemann''s adaptive noise'+
            CRLF+'                     shaping method. takes a parameter to set the order of the'+
            CRLF+'                     FIR filter, (32<=n<=96; default=64; multiple of 8 only);'+
            CRLF+'                     "OFF" to disable adaptive shaping; "NOWARP" to disable '+
            CRLF+'                     default frequency warping;'+
            CRLF+'-a, --analyses <n>   set number of FFT analysis lengths, (2<=n<=6; default=3,'+
            CRLF+'                     i.e. 32, 64 & 1024 samples. n=2, remove 32 sample FFT;'+
            CRLF+'                     n>3 add 512; n>4, add 256; n>6, add 128) nb. FFT lengths.'+
            CRLF+'                     stated are for 44.1/48kHz audio, higher sample rates will'+
            CRLF+'                     automatically increase all FFT lengths as required.'+
            CRLF+'-l, --limit <n>      set upper frequency limit to be used in analyses to n Hz;'+
            CRLF+'                     (10000<=n<=20000; default=16000).'+
            CRLF+'    --linkchannels   revert to original single bits-to-remove value for all'+
            CRLF+'                     channels rather than channel dependent bits-to-remove.'+
            CRLF+'    --maxclips <n>   set max. number of acceptable clips per channel per block;'+
            CRLF+'                     (0<=n<=16; default=3,3,3,3,3,2,2,2,2,2,1,1,1,0,0,0).'+
            CRLF+'-m, --midside        analyse 2 channel audio for mid/side content.'+
            CRLF+'    --nodccorrect    disable DC correction of audio data prior to FFT analysis,'+
            CRLF+'                     default=on; (DC offset calculated per FFT data set).'+
            CRLF+'    --scale <n>      factor to scale audio by; (0.0625<n<=8.0; default=1).'+
            CRLF+'-s, --shaping [n]    enable fixed noise shaping, takes optional parameter [n]'+
            CRLF+'                     to allow user defined shaping proportion (0.0<=n<=1.0),'+
            CRLF+'                     otherwise default to quality setting dependent value.'+
            CRLF+'                     Disables adaptive noise shaping.'+
            CRLF+'    --static <n>     set minimum-bits-to-keep-static to n bits (default=6;'+
            CRLF+'                     7<=n<=28, limited to bits-per-sample - 4).'+
            CRLF+'-U, --underlap <n>   enable underlap mode to increase number of FFT analyses'+
            CRLF+'                     performed at each FFT length, (n = 2, 4 or 8, default=2).'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Output Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'    --bitdist        show distrubution of bits to remove.'+
            CRLF+'    --blockdist      show distribution of lowest / highest significant bit of'+
            CRLF+'                     input codec-blocks and bit-removed codec-blocks.'+
            CRLF+'-d, --detail         enable per block per channel bits-to-remove data display.'+
            CRLF+'-F, --freqdist       enable frequency analysis display of input data.'+
            CRLF+'-H, --histogram      show sample value histogram (input, lossy and correction).'+
            CRLF+'    --longdist       show long frequency distribution data (input/lossy/lwcdf).'+
            CRLF+'    --perchannel     show selected distribution data per channel.'+
            CRLF+'-p, --postanalyse    enable frequency analysis display of output and'+
            CRLF+'                     correction data in addition to input data.'+
            CRLF+'    --sampledist     show distribution of lowest / highest significant bit of'+
            CRLF+'                     input samples and bit-removed samples.'+
            CRLF+'    --spread [full]  show detailed [more detailed] results from the spreading/'+
            CRLF+'                     averaging algorithm.'+
            CRLF+'-W, --width <n>      select width of output options (79<=n<=255).'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'System Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-B, --below          set process priority to below normal.'+
            CRLF+'    --low            set process priority to low.'+
            CRLF+'-N, --nowarnings     suppress lossyWAV warnings.'+
            CRLF+'-Q, --quiet          significantly reduce screen output.'+
            CRLF+'-S, --silent         no screen output.'+
            CRLF;//2345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_special_thanks =
            CRLF+'Special thanks go to:'+CRLF+
            CRLF+'David Robinson       for the publication of his lossyFLAC method, guidance, and'+
            CRLF+'                     the motivation to implement his method as lossyWAV.'+CRLF+
            CRLF+'Horst Albrecht       for ABX testing, valuable support in tuning the internal'+
            CRLF+'                     presets, constructive criticism and all the feedback.'+CRLF+
            CRLF+'Sebastian Gesemann   for the adaptive noise shaping method and the amount of'+
            CRLF+'                     help received in implementing it and also for the basis of'+
            CRLF+'                     the fixed noise shaping method.'+CRLF+
            CRLF+'Matteo Frigo and     for libfftw3-3.dll contained in the FFTW distribution'+
            CRLF+'Steven G Johnson     (v3.2.1 or v3.2.2).'+CRLF+
            CRLF+'Mark G Beckett       for the Delphi unit that provides an interface to the'+
            CRLF+'(Univ. of Edinburgh) relevant fftw routines in libfftw3-3.dll.'+CRLF+
            CRLF+'Don Cross            for the Complex-FFT algorithm originally used.'+
            CRLF;//2345678901234567890123456789012345678901234567890123456789012345678901234567890

  num_quality_synonyms = 6;
var
  quality_synonyms_vals     : array[0..num_quality_synonyms] of double     = (-5.0,-2.5, 0.0, 2.5, 5.0, 7.5,10.0);
  quality_synonyms_short    : array[0..num_quality_synonyms] of string[3]  = ('X','P','C','S','H','E','I');
  quality_synonyms_long     : array[0..num_quality_synonyms] of string[13] = ('extraportable','portable','economic','standard','high','extreme','insane');

  this_quality_synonym_long : string;
  current_parameter         : string;
  parmError                 : string;
  ThisParameterNumber       : Integer; // internal param counter for nextParamStr
  max_open_retries          : integer = 0;

  Unit_Data_Packer2         : array[0..1] of integer;

  //============================================================================
  function parmchar : char;
  //============================================================================
  Begin
  //============================================================================
    parmchar:='@';
    if (length(current_parameter)=2) and (current_parameter[1]='-') then
      parmchar:=current_parameter[2];
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure parmsError;
  //============================================================================
  Begin
  //============================================================================
    if (paramcount>0) and (Parameters.help=0) then
      write(ErrOutput,parmerror,CRLF)
    else
    Begin
      if Parameters.help>1 then
        write(ErrOutput,lossyWAV_process_description+CRLF);
      write(ErrOutput,lossyWAV_standard_help);
      if Parameters.help>1 then
        write(ErrOutput,lossyWAV_advanced_help);
      write(ErrOutput,lossyWAV_Special_Thanks);
    End;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function GetNextParamStr: longbool;
  //============================================================================
  Begin
  //============================================================================
    GetNextParamStr:=false;
    if ThisParameterNumber<paramCount then
    Begin
      ThisParameterNumber:=ThisParameterNumber + 1;
      current_parameter:=ParamStr(ThisParameterNumber);
      GetNextParamStr:=true;
    End;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function NextParameterIsParameterOrEnd : Boolean;
  //============================================================================
  var
    NextParamStr : String;
  Begin
  //============================================================================
    NextParameterIsParameterOrEnd:=False;

    If ThisParameterNumber=ParamCount then
      NextParameterIsParameterOrEnd:=True
    else
    Begin
      NextParamStr:=ParamStr(ThisParameterNumber+1);
      if (NextParamStr[1]='-') then
        if length(NextParamStr)=1 then
          NextParameterIsParameterOrEnd:=True
        else
          if (Ord(NextParamStr[2])>57) or ((Ord(NextParamStr[2])<48) and (Ord(NextParamStr[2])<>46)) then
            NextParameterIsParameterOrEnd:=True;
    End;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure parmerror_multiple_selection;
  //============================================================================
  Begin
  //============================================================================
    lossyWAVerror('Multiple '+parmerror+' switches given.',$31);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure parmerror_mutually_incompatible_selection;
  //============================================================================
  Begin
  //============================================================================
    lossyWAVerror('Switches '+parmerror+' are incompatible.',$31);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure parmerror_no_value_given;
  //============================================================================
  Begin
  //============================================================================
    lossyWAVerror('No '+parmerror+' value given.',$31);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure parmerror_val_error;
  //============================================================================
  Begin
  //============================================================================
    lossyWAVerror('Error evaluating '+parmerror+' value given.',$31);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Integer); Overload;
  //============================================================================
  Begin
  //============================================================================
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+inttostr(cpv_low)+'<=n<='+inttostr(cpv_high),$31);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Double;digits : Integer); Overload;
  //============================================================================
  Begin
  //============================================================================
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+floattostrf(cpv_low,ffFixed,99,digits)+'<=n<='+floattostrf(cpv_high,ffFixed,99,digits),$31);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function check_parameter : boolean;
  //============================================================================
  var
    qs_i, valError : Integer;
  Begin
  //============================================================================
    check_parameter:=false;

    if (parmchar='v') or (current_parameter='--version') then
    Begin
      write(ErrOutput,version_string,strings.version_short,CRLF);
      halt(0);
    End;

    if (parmchar='h') or (current_parameter='--help') then
    Begin
      write(ErrOutput,version_string,strings.version_short,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      Parameters.help:=1;
      parmserror;
      halt(0);
    End;

    if (parmchar='L') or (current_parameter='--longhelp') then
    Begin
      write(ErrOutput,version_string,strings.version_short,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      Parameters.help:=2;
      parmserror;
      halt(0);
    End;

    if (parmchar='q') or (current_parameter='--quality') then
    Begin
      parmerror:='quality preset';

      if Parameters.quality<>-99 then
        parmerror_multiple_selection;

      if not GetNextParamStr then
        parmerror_no_value_given;

      for qs_i:=0 to num_quality_synonyms do
      Begin
        this_quality_synonym_long:=quality_synonyms_long[qs_i];
        if (current_parameter=quality_synonyms_short[qs_i]) or (current_parameter=this_quality_synonym_long) then
        Begin
          parmerror:='quality preset';
          Parameters.quality:=quality_synonyms_vals[qs_i];
          Strings.Parameter:='--quality '+this_quality_synonym_long;
          check_parameter:=true;
          exit
        End;
      End;
      if Parameters.quality=-99 then
      Begin
        Val(current_parameter, Parameters.quality, valError);
        if valError <> 0 then
          parmerror_val_error;
        Parameters.quality:=nRound10(Parameters.quality,4);
        check_permitted_values(Parameters.quality,-5,10,4);
        if Parameters.quality=int(Parameters.quality) then
          Strings.Parameter:='--quality '+inttostr(floor(Parameters.quality))
        else
          Strings.Parameter:='--quality '+floattostrf(Parameters.quality,ffFixed,99,4);
        check_parameter:=true;
        exit;
      End;
    End;

    if (current_parameter='--maxclips') then
    Begin
      parmerror:='maximum clips';
      if Parameters.maxclips<>-1 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.maxclips, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.maxclips,0,16);
      check_parameter:=true;
      exit;
    End;

    if (parmchar='-') or (current_parameter='--stdout') then
    Begin
      parmerror:='output to STDOUT';
      if Parameters.STDOUTPUT=true then
        parmerror_multiple_selection;
      Parameters.STDOUTPUT:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='B') or (current_parameter='--below') then
    Begin
      parmerror:='process priority';
      if (Parameters.priority>0) then
        parmerror_multiple_selection;
      setpriority($00004000);
      Parameters.priority:=1;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--low') then
    Begin
      parmerror:='process priority';
      if (Parameters.priority>0) then
        parmerror_multiple_selection;
      Parameters.priority:=2;
      setpriority($00000040);
      check_parameter:=true;
      exit;
    End;

    if (parmchar='d') or (current_parameter='--detail') then
    Begin
      parmerror:='detail';
      if (Parameters.detail) then
        parmerror_multiple_selection;
      Parameters.detail:=true;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--perchannel') then
    Begin
      parmerror:='per-channel data output';
      if (Parameters.perchannel) then
        parmerror_multiple_selection;
      Parameters.perchannel:=true;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--bitdist') then
    Begin
      parmerror:='bits-to-remove distribution';
      if (Parameters.bitdist) then
        parmerror_multiple_selection;
      Parameters.bitdist:=true;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--spread') then
    Begin
      parmerror:='spreading results';
      if (Parameters.Spread<>-1) then
        parmerror_multiple_selection;
      if NextParameterIsParameterOrEnd then
        Parameters.Spread:=1
      else
      Begin
        GetNextParamStr;
        If not (Current_Parameter='full') then
        Begin
          ParmError:='spreading results';
          parmerror_val_error;
        End
        Else
          Parameters.Spread:=2;
      End;

      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--blockdist') then
    Begin
      parmerror:='block lsb distribution';
      if (Parameters.blockdist) then
        parmerror_multiple_selection;
      Parameters.blockdist:=true;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--sampledist') then
    Begin
      parmerror:='sample lsb distribution';
      if (Parameters.sampledist) then
        parmerror_multiple_selection;
      Parameters.sampledist:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='F') or (current_parameter='--freqdist') then
    Begin
      parmerror:='input frequency distribution';
      if (Parameters.freqdist) then
        parmerror_multiple_selection;
      Parameters.freqdist:=true;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--longdist') then
    Begin
      parmerror:='long frequency distribution';
      if (Parameters.longdist) then
        parmerror_multiple_selection;
      Parameters.longdist:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='W') or (current_parameter='--width') then
    Begin
      parmerror:='wide output';
      if Parameters.width<>-1 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter,Parameters.width, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.width,79,255);
      check_parameter:=true;
      exit;
    End;

    if (parmchar='H') or (current_parameter='--histogram') then
    Begin
      parmerror:='sample value histogram';
      if (Parameters.histogram) then
        parmerror_multiple_selection;
      Parameters.histogram:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='p') or (current_parameter='--postanalyse') then
    Begin
      parmerror:='output frequency distribution';
      if (Parameters.postanalyse) then
        parmerror_multiple_selection;
      if not parameters.freqdist then
      begin
        lossywavwarning('freqdist not selected before postanalyse');
        parameters.freqdist:=true;
      end;
      Parameters.postanalyse:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='Q') or (current_parameter='--quiet') then
    Begin
      parmerror:='reduced output';
      if not Parameters.verbosity then
        parmerror_multiple_selection;
      Parameters.verbosity:=false;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='S') or (current_parameter='--silent') then
    Begin
      parmerror:='silent running';
      if Parameters.silent=true then
        parmerror_multiple_selection;
      Parameters.silent:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='N') or (current_parameter='--nowarnings') then
    Begin
      parmerror:='no warnings';
      if not Parameters.warnings then
        parmerror_multiple_selection;
      Parameters.warnings:=false;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='c') or (current_parameter='--check') then
    Begin
      parmerror:='check';
      if Parameters.checking then
        parmerror_multiple_selection;
      Parameters.checking:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='M') or (current_parameter='--merge') then
    Begin
      parmerror:='merge';
      if Parameters.merging then
        parmerror_multiple_selection;
      Parameters.merging:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='w') or (current_parameter='--writetolog') then
    Begin
      parmerror:='write log output';
      if Parameters.writetolog then
        parmerror_multiple_selection;
      Parameters.writetolog:=true;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--stdinname') then
    Begin
      parmerror:='pseudo filename';
      if Parameters.stdinname<>'' then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Parameters.stdinname:=current_parameter;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='f') or (current_parameter='--force') then
    Begin
      parmerror:='forced over-write';
      if Parameters.forcing then
        parmerror_multiple_selection;
      Parameters.forcing:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='o') or (current_parameter='--outdir') then
    Begin
      parmerror:='destination folder';
      if (Parameters.WAVOutDir <> '') then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Parameters.WAVOutDir:=current_parameter;
      if not((Parameters.WAVOutDir[length(Parameters.WAVOutDir)]='\') or (Parameters.WAVOutDir[length(Parameters.WAVOutDir)]=':')) then
        Parameters.WAVOutdir:=Parameters.WAVOutdir+'\';
      check_parameter:=true;
      exit;
    End;

    if (parmchar='C') or (current_parameter='--correction') then
    Begin
      parmerror:='correction file';
      if Parameters.correction then
        parmerror_multiple_selection;
      Parameters.correction:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='s') or (current_parameter='--shaping') then
    Begin
      parmerror:='adaptive and fixed noise shaping';
      if parameters.adaptive_shaping<>-2 then
        parmerror_mutually_incompatible_selection;
      Parameters.Adaptive_Shaping:=-3;
      Parameters.Adaptive_Shaping_Warp:=false;

      parmerror:='fixed noise shaping';
      if Parameters.shaping<>-1 then
        parmerror_multiple_selection;

      if not nextparameterisparameterorend then
      Begin
        GetNextParamStr;
        Val(current_parameter, Parameters.shaping, valError);
        if valError <> 0 then
          parmerror_val_error;
        Parameters.shaping:=nRound10(Parameters.shaping,4);
        check_permitted_values(Parameters.shaping,0,1,4);
        End
      else
        Parameters.shaping:=-2;

      check_parameter:=true;
      exit;
    End;

    if (parmchar='A') or (current_parameter='--adaptive') then
    Begin
      parmerror:='fixed and adaptive noise shaping';
      if parameters.shaping<>-1 then
        parmerror_mutually_incompatible_selection;

      parmerror:='adaptive noise shaping';

      if nextparameterisparameterorend then
        parmerror_no_value_given
      else
      repeat
        GetNextParamStr;

        if ((Current_Parameter='NOWARP') or (Current_Parameter='nowarp'))then
        Begin
          if parameters.adaptive_shaping=-1 then
          begin
            parmerror:='adaptive off and nowarp';
            parmerror_mutually_incompatible_selection
          end;
          if parameters.adaptive_shaping_warp then
            parameters.adaptive_shaping_warp:=false
          else
            parmerror_multiple_selection;
        End
        else
          if (current_parameter='OFF') or (current_parameter='off') then
          Begin
            if not parameters.adaptive_shaping_warp then
            begin
              parmerror:='adaptive off and nowarp';
              parmerror_mutually_incompatible_selection;
            end;
            if parameters.adaptive_shaping=-2 then
              parameters.adaptive_shaping:=-1
            else
              parmerror_multiple_selection;
          End
          else
          Begin
            if parameters.adaptive_shaping<>-2 then
              parmerror_multiple_selection;

            Val(current_parameter, Parameters.adaptive_shaping, valError);
            if valError <> 0 then
              parmerror_val_error;

            check_permitted_values(Parameters.adaptive_shaping,32,96);
            if parameters.adaptive_shaping and 7<>0 then
              lossyWAVerror('Adaptive noise shaping value must be a multiple of 8: '+inttostr(parameters.adaptive_shaping)+' is invalid.',$31);
          End;

      until NextParameterIsParameterOrEnd;
      check_parameter:=true;
      exit;
    End;

    if                   (current_parameter='--static') then
    Begin
      parmerror:='minimum bits to keep (static)';
      if Parameters.static<>STATIC_MINIMUM_BITS_TO_KEEP then
        parmerror_multiple_selection;

      if not GetNextParamStr then
        parmerror_no_value_given;

      Val(current_parameter, Parameters.static, valError);
      if valError <> 0 then
        parmerror_val_error;

      check_permitted_values(Parameters.static,7,28);
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--scale') then
    Begin
      parmerror:='scaling factor';
      if Parameters.scaling<>-1 then
        parmerror_multiple_selection;

      if not GetNextParamStr then
        parmerror_no_value_given;

      Val(current_parameter, Parameters.scaling, valError);
      if valError <> 0 then
        parmerror_val_error;

      Parameters.scaling:=nRound10(Parameters.scaling,5);
      check_permitted_values(Parameters.scaling,0.0625,8.0,5);
      check_parameter:=true;
      exit;
    End;

    if (parmchar='l') or (current_parameter='--limit') then
    Begin
      parmerror:='limit frequency';
      if Parameters.limit<>-1 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.limit, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.limit,10000,20000);
      check_parameter:=true;
      exit;
    End;

    if (parmchar='a') or (current_parameter='--analyses') then
    Begin
      parmerror:='number of FFT analyses';
      if Parameters.analyses<>3 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.analyses, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.analyses,2,6);
      check_parameter:=true;
      exit;
    End;

    if (parmchar='m') or (current_parameter='--midside') then
    Begin
      parmerror:='midside';
      if Parameters.midside then
        parmerror_multiple_selection;
      Parameters.midside:=true;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--nodccorrect') then
    Begin
      parmerror:='DC correction';
      if not Parameters.dccorrect then
        parmerror_multiple_selection;
      Parameters.dccorrect:=false;
      check_parameter:=true;
      exit;
    End;

    if (current_parameter='--linkchannels') then
    Begin
      parmerror:='link channels';
      if Parameters.linkchannels then
        parmerror_multiple_selection;
      Parameters.linkchannels:=true;
      check_parameter:=true;
      exit;
    End;

    if (parmchar='U') or (current_parameter='--underlap') then
    Begin
      parmerror:='underlap';
      if Parameters.underlap<>0 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.underlap, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.underlap,2,8);
      if (Parameters.underlap and (parameters.underlap-1))>0 then
        lossyWAVerror('Permitted underlap values : 2, 4, 8',$31);
      check_parameter:=true;
      exit;
    End;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function getParms: longbool;
  //============================================================================
  var
    colonpos : Integer;
    slashpos : Integer;
  Begin
  //============================================================================
    getParms        :=false;

    ThisParameterNumber:=0;

    Parameters.WAVInpDir        :='';
    Parameters.WAVOutDir        :='';
    Parameters.wavName          :='';
    Parameters.stdinname        :='';

    Parameters.priority         :=0;
    Parameters.quality          :=-99;
    Parameters.scaling          :=-1;
    Parameters.shaping          :=-1;

    Parameters.adaptive_shaping       :=-2;
    Parameters.adaptive_shaping_warp  :=true;

    Parameters.analyses         :=3;
    Parameters.maxclips         :=-1;
    Parameters.help             :=0;

    Parameters.limit            :=-1;
    Parameters.underlap         :=0;

    Parameters.static           :=STATIC_MINIMUM_BITS_TO_KEEP;
    Parameters.dynamic          :=-1;

    Parameters.perchannel   :=false;
    Parameters.bitdist      :=false;
    Parameters.Spread       :=-1;
    Parameters.sampledist   :=false;
    Parameters.Spread       :=-1;
    Parameters.freqdist     :=false;
    Parameters.longdist     :=false;
    Parameters.histogram    :=false;
    Parameters.blockdist    :=false;
    Parameters.checking     :=false;
    Parameters.correction   :=false;
    Parameters.width        :=-1;

    Parameters.detail       :=false;
    Parameters.forcing      :=false;
    Parameters.linkchannels :=false;

    Parameters.merging      :=false;
    Parameters.midside      :=false;
    Parameters.dccorrect    :=true;
    Parameters.sampledist   :=false;
    Parameters.silent       :=false;

    Parameters.STDINPUT     :=false;
    Parameters.STDOUTPUT    :=false;
    Parameters.warnings     :=true;
    Parameters.writetolog   :=false;

    Parameters.verbosity    :=true;

    if not GetNextParamStr then
      exit;

    if (current_parameter='-') then
      Parameters.STDINPUT:=true
    else
      if (current_parameter[1]='-') then
      Begin
        if (parmchar='v') or (current_parameter='--version') then
        Begin
          write(ErrOutput,version_string,strings.version_short,CRLF);
          halt(0);
        end
        else
        Begin
          write(ErrOutput,version_string,strings.version_short,lossyWAVhead1,CRLF);
          write(ErrOutput,lossyWAV_GPL,CRLF);
          if (parmchar='h') or (current_parameter='--help') or (parmchar='L') or (current_parameter='--longhelp') then
          Begin
            Parameters.help:=1+ord((current_parameter='--longhelp') or (parmchar='L'));
            parmserror;
            halt(0);
          end
          else
            lossyWAVerror('No input file given.',1);
        End;
      end
      else
        Parameters.wavName:=current_parameter;

    if (Parameters.wavname<>'') then
    Begin
      if (Parameters.wavName[1]='"') then
        Parameters.wavName:=copy(Parameters.wavName,2,length(Parameters.wavName)-1);
      if (Parameters.wavName[length(Parameters.wavName)]='"') then
        Parameters.wavName:=copy(Parameters.wavName,1,length(Parameters.wavName)-1);

      colonpos:=pos(':',Parameters.wavName);
      if colonpos>0 then
      Begin
        Parameters.wavInpDir:=copy(Parameters.wavName,1,colonpos);
        Parameters.wavName:=copy(Parameters.wavName,colonpos+1,length(Parameters.wavName)-colonpos);
      End;

      slashpos:=pos('\',Parameters.wavName);
      while slashpos>0 do
      Begin
        Parameters.wavInpDir:=Parameters.wavInpDir+copy(Parameters.wavName,1,slashpos);
        Parameters.wavName:=copy(Parameters.wavName,slashpos+1,length(Parameters.wavName)-slashpos);
        slashpos:=pos('\',Parameters.wavName);
      End;
    End;

    while GetNextParamStr do
    Begin
      if not check_parameter then
      Begin
        parmError:='%lossyWAV Error% : Incorrect option: "' + current_parameter + '"'+CRLF;
        exit;
      End;
    End;
    getParms:=true;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Function wavFilePrintName : String;
  //============================================================================
  Begin
  //============================================================================
    if Parameters.STDINPUT then
    Begin
      if Parameters.stdinname='' then
        wavFilePrintName:='from STDIN'
      else
        wavfileprintname:=Parameters.stdinname;
    end
    else
      wavFilePrintName:=Parameters.wavname;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function wavInp: string;
  //============================================================================
  Begin
  //============================================================================
    wavInp:=Parameters.wavInpDir + Parameters.wavName;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Function WAVOut: String;
  //============================================================================
  Begin
  //============================================================================
    WAVOut:=Parameters.WAVOutDir + Parameters.WAVName;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function lossyInp: String;
  //============================================================================
  Begin
  //============================================================================
    lossyInp:=Parameters.WAVInpDir + Parameters.lossyName;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function lossyOut: string;
  //============================================================================
  Begin
  //============================================================================
    lossyOut:=Parameters.WAVOutDir + Parameters.lossyName;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function lwcdfInp: String;
  //============================================================================
  Begin
  //============================================================================
    lwcdfInp:=Parameters.WAVInpDir + Parameters.lwcdfName;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function lwcdfOut: string;
  //============================================================================
  Begin
  //============================================================================
    lwcdfOut:=Parameters.WAVOutDir + Parameters.lwcdfName;
  //============================================================================
  End;
  //============================================================================

  {$I-}
  //============================================================================
  procedure open_log_file;
  //============================================================================
  var
    open_retries : Integer;
    lastioresult : Integer;
  Begin
  //============================================================================
    Open_Retries:=0;
    if not Parameters.logfileopened then
    Begin
      filemode:=1;                                                  //write only
      assign(LogOutput,Parameters.logfilename);
      if fileexists(Parameters.logfilename) then
      Begin
        if filegetattr(Parameters.logfilename) and 1=1 then
        Begin
          lossyWAVwarning('lossyWAV.log file is read-only.');
          if(filesetattr(Parameters.logfilename,filegetattr(Parameters.logfilename) and $FE)<>0) then
            lossyWAVerror('Cannot gain write access to '+Parameters.logfilename,$31);
        End;
        Repeat
          {$I-}
          Append(LogOutput);
          lastIOResult:=ioresult;
          {$I+}
          if lastioresult<>0 then
          Begin
            inc(open_retries);
            timer.DelayTime:=PowersOf.two[-5];
            ndelay;
          End;
        Until (lastioresult=0) or (open_retries=128);
        if lastioresult<>0 then
          lossywaverror('Opening lossyWAV.log',$31);
        max_open_retries:=max(open_retries,max_open_retries);
      end
      else
      Begin
        {$I-}
        Rewrite(LogOutput);
        lastioresult:=ioresult;
        {$I+}
        if lastioresult<>0 then
          lossywaverror('Opening lossyWAV.log',$31);
      End;
    End;
    Parameters.logfileopened:=true;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure close_log_file;
  //============================================================================
  Begin
  //============================================================================
    if Parameters.logfileopened then
    Begin
      flush(LogOutput);
      close(LogOutput);
    End;
    Parameters.logfileopened:=false;
  //============================================================================
  End;
  //============================================================================
  {$I+}

  //============================================================================
  procedure nParameter_Init;
  //============================================================================
  Begin
  //============================================================================
    Parameters.parameters_checked:=false;
    Parameters.logfileopened:=false;

    {$I-}
    if (not getparms) then
    Begin
      write(ErrOutput,version_string,strings.version_short,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      if parmerror<>'' then
        write(ErrOutput,parmerror);
      write(ErrOutput,'Syntax: lossyWAV <wavfilename>'+CRLF);
      write(ErrOutput,CRLF+'Type ''lossyWAV --help'' for more detail.'+CRLF);
      halt(1);
    End;

    if Parameters.quality=-99 then
    Begin
      Parameters.quality:=2.5;
      Strings.Parameter:='--quality standard';
    End;

    if (Parameters.help>=1) then
    Begin
      write(ErrOutput,version_string,strings.version_short,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      parmserror;
      halt(1);
    End;

    if (Parameters.verbosity) and (not Parameters.silent) then
      write(ErrOutput,version_string,strings.version_short,lossyWAVhead1,lossywavhead2);

    if (not Parameters.STDINPUT) and (length(Parameters.wavName)=0) then
      lossyWAVerror('Name of input or output wav file missing.',$31);

    if (Parameters.STDINPUT) then
    Begin
      Parameters.wavname:='';
      Parameters.lossyName:='lossywav.lossy.wav';
      Parameters.lwcdfname:='lossywav.lwcdf.wav';
      if (Parameters.merging) then
        lossyWAVerror('Merge parameter is incompatible'+CRLF+'                   with STDIN file input mode.',$31);
    end
    else
    Begin
      if not fileexists(wavInp) then
      Begin
        if not fileexists(wavInp+'.wav') then
          lossyWAVerror('Input file: '+Parameters.wavName+' does not exist.',$31)
        else
          Parameters.wavName:=Parameters.wavName+'.wav';
      End;
    End;

    if (Parameters.STDOUTPUT=true) then
    Begin
      Parameters.lossyname:='';
      Parameters.lwcdfname:='';
      Parameters.wavoutdir:='';
      if (Parameters.correction) then
        lossyWAVerror('Correct parameter is incompatible'+CRLF+'                   with STDOUT file output mode.',$31);
    end
    else
    Begin
      if Parameters.stdinput then
      Begin
        if Parameters.stdinname<>'' then
        Begin
          if (length(Parameters.stdinname)<5) then
            if copy(Parameters.stdinname,length(Parameters.stdinname),1)='.' then
              Parameters.stdinname:=Parameters.stdinname+'wav'
            else
              Parameters.stdinname:=Parameters.stdinname+'.wav';

          Parameters.lossyName:=copy(Parameters.stdinname, 1, length(Parameters.stdinname) - 3) + 'lossy.' + copy(Parameters.stdinname, length(Parameters.stdinname) - 2, 3);
          Parameters.lwcdfName:=copy(Parameters.stdinname, 1, length(Parameters.stdinname) - 3) + 'lwcdf.' + copy(Parameters.stdinname, length(Parameters.stdinname) - 2, 3);
        end
        else
        Begin
          Parameters.lossyname:='stdin.lossy.wav';
          Parameters.lwcdfname:='stdin.lwcdf.wav';
        End;
      end
      else
      Begin
        Parameters.lossyName:=copy(Parameters.wavName, 1, length(Parameters.wavName) - 3) + 'lossy.' + copy(Parameters.wavName, length(Parameters.wavName) - 2, 3);
        Parameters.lwcdfName:=copy(Parameters.wavName, 1, length(Parameters.wavName) - 3) + 'lwcdf.' + copy(Parameters.wavName, length(Parameters.wavName) - 2, 3);
      End;

      if (Parameters.WAVOutdir<>'') and (directoryexists(Parameters.WAVOutdir)=false) then
        lossyWAVerror('Directory : '+Parameters.WAVOutdir+' cannot be accessed.',$31);
    End;

    if not Parameters.merging then
    Begin
      if (Parameters.STDOUTPUT=false) and (fileexists(lossyOut)=true) then
      Begin
        if Parameters.forcing=true then
        Begin
          if (filegetattr(lossyOut) and 1)=1 then
          Begin
            lossyWAVwarning('Output lossy file is read-only.');
            if(filesetattr(lossyOut,filegetattr(lossyOut) and $FE)<>0) then
              lossyWAVerror('Cannot gain write access to '+lossyOut,$31);
          End;
        End
        else
          lossyWAVerror('Output file: '+lossyOut+' exists.',$31);
      End;

      if (Parameters.correction) and (fileexists(lwcdfOut)=true) then
      Begin
        if (Parameters.forcing=true) then
        Begin
          if filegetattr(lwcdfOut) and 1=1 then
          Begin
            lossyWAVwarning('Output correction file is read-only.');
            if(filesetattr(lwcdfOut,filegetattr(lwcdfOut) and $FE)<>0) then
              lossyWAVerror('Cannot gain write access to '+lwcdfOut,$31);
          End;
        End
        else
          lossyWAVerror('Output file: '+lwcdfOut+' exists.',$31);
      End;
    end
    else
    Begin
      Parameters.lossyName:=Parameters.wavName;
      if (copy(Parameters.lossyName,length(Parameters.lossyName)-9,10)='.lossy.wav') then
      Begin
        Parameters.lwcdfName:=copy(Parameters.wavName,1,length(Parameters.wavName)-10)+'.lwcdf.wav';
        Parameters.wavName:=copy(Parameters.wavName,1,length(Parameters.wavName)-10)+'.wav';
      End;

      if fileexists(WAVOut) and (not parameters.forcing) then
        lossyWAVerror('Output merged file already exists',$31);

      if not fileexists(lossyInp) then
        lossyWAVerror('Input lossy file not found.',$31);

      if not fileexists(lwcdfInp) then
        lossyWAVerror('Input correction file not found.',$31);
    End;

    Parameters.logfilename:=Parameters.WAVOutDir+'lossyWAV.log';
    {$I+}
   //============================================================================
  End;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  nParameter_Init;
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================

