//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nProcess;
//==============================================================================
interface
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================

Procedure Process_This_Codec_Block;

Procedure nProcess_Init;

//==============================================================================
implementation
//==============================================================================
uses
  nCore,
  fftw_interface,
  {$IFDEF USEASM}
  nFFT_asm,
  {$ELSE}
  nFFT_pas,
  {$ENDIF}
  nParameter,
  nFillFFT,
  nMaths,
  nOutput,
  nRemoveBits,
  nSGNS,
  nSpreading,
  math,
  SysUtils;
//==============================================================================
const
  UNITNAME='nProcess';
var
  {$A4}
  Unit_Data_Packer2 : Record
    Data : Array[0..2] of Integer;
  End;
  {$A8}

  {$IFNDEF USEASM}
  Procedure Fill_Last_from_SGNS;
  Var
    sa_i : Integer;
    temp_result : Double;
  Begin
    for sa_i:=0 to Global.FFT.Length_Half do
    Begin
      temp_result:=SGNSFFT.Last_Results[Global.Channel,sa_i];
      fft_last_result[sa_i]:=temp_result;
      fft_raw_result[sa_i]:=fft_raw_result[sa_i]+temp_result;
    End;
  End;
  {$ELSE}
  Procedure Fill_Last_from_SGNS;
  asm
                push    eax
                push    ebx
                push    edx

                mov     eax,Global.Channel
                mov     ebx,dword ptr SGNSFFT.Channel_Offset[eax*INTEGER_SIZE]
                xor     eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$ENDIF}

      @@01:     fld     qword ptr SGNSFFT.Last_Results[ebx+eax*DOUBLE_SIZE]
                fst     qword ptr fft_last_result[eax*DOUBLE_SIZE]
                fadd    qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fstp    qword ptr fft_raw_result[eax*DOUBLE_SIZE]

                add     eax,1
                cmp     eax,Global.FFT.Length_Half
                jle     @@01

                pop     edx
                pop     ebx
                pop     eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Fill_Last_from_History;
  Var
    sa_i : Integer;
    temp_result : Double;
  Begin
    for sa_i:=0 to Global.FFT.Length_Half do
    Begin
      temp_result:=History.Last_Results[Global.Channel,sa_i];
      fft_last_result[sa_i]:=temp_result;
      fft_raw_result[sa_i]:=fft_raw_result[sa_i]+temp_result;
    End;
  End;
  {$ELSE}
  Procedure Fill_Last_from_History;
  asm
                push    eax
                push    ebx
                push    edx

                mov     eax,Global.Channel
                mov     ebx,dword ptr History.Channel_Offset[eax*INTEGER_SIZE]
                xor     eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$ENDIF}

      @@01:     fld     qword ptr History.Last_Results[ebx+eax*DOUBLE_SIZE]
                fst     qword ptr fft_last_result[eax*DOUBLE_SIZE]
                fadd    qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fstp    qword ptr fft_raw_result[eax*DOUBLE_SIZE]

                add     eax,1
                cmp     eax,Global.FFT.Length_Half
                jle     @@01

                pop     edx
                pop     ebx
                pop     eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Fill_History_from_Last;
  Var
    sa_i : Integer;
    temp_result : double;
  Begin
    for sa_i:=0 to Global.FFT.length_half do
    Begin
      temp_result:=fft_last_result[sa_i];
      History.Last_Results[Global.Channel,sa_i]:=temp_result;
      temp_result:=OneOver[Process.Analyses_Performed-1]*(fft_raw_result[sa_i]-temp_result);
      history.DATA_results[Global.Channel,sa_i]:=history.DATA_results[Global.Channel,sa_i]+temp_result;

      SGNSFFT.FFT_Results_Short[sa_i]:=temp_result;
    End;
  End;
  {$ELSE}
  Procedure Fill_History_from_Last;
  asm
                push    eax
                push    ebx
                push    edx

                mov     eax,Process.Analyses_Performed
                fld     qword ptr OneOver[eax*DOUBLE_SIZE-DOUBLE_SIZE]

                mov     eax,Global.Channel
                mov     ebx,dword ptr History.Channel_Offset[eax*INTEGER_SIZE]
                xor     eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}

      @@01:     fld     qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fld     qword ptr fft_last_result[eax*double_size]
                fst     qword ptr History.Last_Results[ebx+eax*double_size]
                fsub
                fmul    st,st(1)

                fst     qword ptr SGNSFFT.FFT_Results_Short[eax*double_size]
                fadd    qword Ptr History.Data_Results[ebx+eax*double_size]
                fstp    qword ptr History.Data_Results[ebx+eax*double_size]

                add     eax,1
                cmp     eax,Global.FFT.Length_Half
                jle     @@01

                fstp    st

                pop     edx
                pop     ebx
                pop     eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Fill_SGNSFFT_from_Last;
  Var
    sa_i : Integer;
    temp_result : Double;
  Begin
    for sa_i:=0 to Global.FFT.Length_Half do
    Begin
      temp_result:=fft_last_result[sa_i];
      SGNSFFT.Last_Results[Global.Channel,sa_i]:=temp_result;
      SGNSFFT.FFT_Results_Long[sa_i]:=fft_raw_result[sa_i]-temp_result;
    End;
  End;
  {$ELSE}
  Procedure Fill_SGNSFFT_from_Last;
  asm
                push    eax
                push    ebx
                push    edx

                mov     eax,Global.Channel
                mov     ebx,dword ptr SGNSFFT.Channel_Offset[eax*INTEGER_SIZE]
                xor     eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@01:     fld     qword ptr fft_raw_result[eax*double_size]
                fld     qword ptr fft_last_result[eax*double_size]
                fst     qword ptr SGNSFFT.Last_Results[ebx+eax*double_size]
                fsub
                fstp    qword ptr SGNSFFT.FFT_Results_Long[eax*double_size]

                add     eax,1
                cmp     eax,Global.FFT.Length_Half
                jle     @@01

                pop     edx
                pop     ebx
                pop     eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Fill_Last_with_Zero;
  Var
    sa_i : Integer;
  Begin
    for sa_i:=0 to Global.FFT.Length_Half do
    Begin
      fft_last_result[sa_i]:=0;
      FFT_skew_Result[sa_i]:=0;
    End
  End;
  {$ELSE}
  Procedure Fill_Last_with_Zero;
  Asm
                push      eax
                push      ecx

                xor       eax,eax
                mov       ecx,Global.FFT.Length_Half

                fldz

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;
                {$ENDIF}

    @@01:       fst       qword ptr fft_last_result[eax*double_size]
                add       eax,1
                cmp       eax,ecx
                jle       @@01

                fstp      st

                pop       ecx
                pop       eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Zero_FFT_Raw_Results;
  Var
    sa_i : Integer;
  Begin
    for sa_i:=0 to Global.FFT.Length_Half do
      fft_raw_result[sa_i]:=0;
  End;
  {$ELSE}
  Procedure Zero_FFT_Raw_Results;
  Asm
                push      eax
                push      ecx

                xor       eax,eax
                mov       ecx,Global.FFT.Length_Half

                fldz

    @@01:       fst       qword ptr fft_raw_result[eax*double_size]
                add       eax,1
                cmp       eax,ecx
                jle       @@01

                fstp      st

                pop       ecx
                pop       eax
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Add_To_BTRD_History;
  Var
    sa_i : Integer;
  Begin
    for sa_i:=0 to Global.FFT.length_half do
      history.BTRD_Results[Global.Channel,sa_i]:=history.BTRD_Results[Global.Channel,sa_i]+(fft_raw_result[sa_i]*OneOver[Process.Analyses_Performed]);
  End;
  {$ELSE}
  Procedure Add_To_BTRD_History;
  Asm
                push      eax
                push      ebx
                push      edx

                mov       eax,Process.Analyses_Performed
                fld       qword ptr OneOver[eax*DOUBLE_SIZE]

                mov       eax,Global.Channel
                mov       ebx,dword ptr History.Channel_Offset[eax*INTEGER_SIZE]
                xor       eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@01:     fld       qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fmul      st,st(1)

                fadd      qword Ptr History.BTRD_Results[ebx+eax*double_size]
                fstp      qword ptr History.BTRD_Results[ebx+eax*double_size]

                add       eax,1
                cmp       eax,Global.FFT.Length_Half
                jle       @@01

                fstp      st

                pop       edx
                pop       ebx
                pop       eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Add_To_CORR_History;
  Var
    sa_i : Integer;
  Begin
    for sa_i:=0 to Global.FFT.length_half do
      history.CORR_Results[Global.Channel,sa_i]:=history.CORR_Results[Global.Channel,sa_i]+(fft_raw_result[sa_i]*OneOver[Process.Analyses_Performed]);
  End;
  {$ELSE}
  Procedure Add_To_CORR_History;
  Asm
                push      eax
                push      ebx
                push      edx

                mov       eax,Process.Analyses_Performed
                fld       qword ptr OneOver[eax*DOUBLE_SIZE]

                mov       eax,Global.Channel
                mov       ebx,dword ptr History.Channel_Offset[eax*INTEGER_SIZE]
                xor       eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@01:     fld       qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fmul      st,st(1)

                fadd      qword Ptr History.CORR_Results[ebx+eax*double_size]
                fstp      qword ptr History.CORR_Results[ebx+eax*double_size]

                add       eax,1
                cmp       eax,Global.FFT.Length_Half
                jle       @@01

                fstp      st

                pop       edx
                pop       ebx
                pop       eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Add_to_DATA_LongDist;
  Var
    sa_i : Integer;
  Begin
    for sa_i:=0 to Global.FFT.length_half do
      LongDist.DATA_Results[Global.Channel,sa_i]:=LongDist.DATA_Results[Global.Channel,sa_i]+(fft_raw_result[sa_i]*OneOver[Process.Analyses_Performed]);
  End;
  {$ELSE}
  Procedure Add_to_DATA_LongDist;
  Asm
                push      eax
                push      ebx
                push      edx

                mov       eax,Process.Analyses_Performed
                fld       qword ptr OneOver[eax*DOUBLE_SIZE]

                mov       eax,Global.Channel
                mov       ebx,dword ptr LongDist.Channel_Offset[eax*INTEGER_SIZE]
                xor       eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@01:     fld       qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fmul      st,st(1)

                fadd      qword Ptr LongDist.DATA_Results[ebx+eax*double_size]
                fstp      qword ptr LongDist.DATA_Results[ebx+eax*double_size]

                add       eax,1
                cmp       eax,Global.FFT.Length_Half
                jle       @@01

                fstp      st

                pop       edx
                pop       ebx
                pop       eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Add_to_BTRD_LongDist;
  Var
    sa_i : Integer;
  Begin
    for sa_i:=0 to Global.FFT.length_half do
      LongDist.BTRD_Results[Global.Channel,sa_i]:=LongDist.BTRD_Results[Global.Channel,sa_i]+(fft_raw_result[sa_i]*OneOver[Process.Analyses_Performed]);
  End;
  {$ELSE}
  Procedure Add_to_BTRD_LongDist;
  Asm
                push      eax
                push      ebx
                push      edx

                mov       eax,Process.Analyses_Performed
                fld       qword ptr OneOver[eax*DOUBLE_SIZE]

                mov       eax,Global.Channel
                mov       ebx,dword ptr LongDist.Channel_Offset[eax*INTEGER_SIZE]
                xor       eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@01:     fld       qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fmul      st,st(1)

                fadd      qword Ptr LongDist.BTRD_Results[ebx+eax*double_size]
                fstp      qword ptr LongDist.BTRD_Results[ebx+eax*double_size]

                add       eax,1
                cmp       eax,Global.FFT.Length_Half
                jle       @@01

                fstp      st

                pop       edx
                pop       ebx
                pop       eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Add_to_CORR_LongDist;
  Var
    sa_i : Integer;
  Begin
    for sa_i:=0 to Global.FFT.length_half do
      LongDist.CORR_Results[Global.Channel,sa_i]:=LongDist.CORR_Results[Global.Channel,sa_i]+(fft_raw_result[sa_i]*OneOver[Process.Analyses_Performed]);
  End;
  {$ELSE}
  Procedure Add_to_CORR_LongDist;
  Asm
                push      eax
                push      ebx
                push      edx

                mov       eax,Process.Analyses_Performed
                fld       qword ptr OneOver[eax*DOUBLE_SIZE]

                mov       eax,Global.Channel
                mov       ebx,dword ptr LongDist.Channel_Offset[eax*INTEGER_SIZE]
                xor       eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@01:     fld       qword ptr fft_raw_result[eax*DOUBLE_SIZE]
                fmul      st,st(1)

                fadd      qword Ptr LongDist.CORR_Results[ebx+eax*double_size]
                fstp      qword ptr LongDist.CORR_Results[ebx+eax*double_size]

                add       eax,1
                cmp       eax,Global.FFT.Length_Half
                jle       @@01

                fstp      st

                pop       edx
                pop       ebx
                pop       eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  Procedure Post_Process_FFT_Array_to_Last;
  Var
    sa_i : Integer;
    sa_x : Double;
  Begin
    for sa_i:=0 to Global.FFT.Length_Half do
    Begin
      with FFT_Array.DComplex[sa_i] do
        sa_x:=(re*re+im*im);
      fft_last_result[sa_i]:=sa_x;
      fft_raw_result[sa_i]:=fft_raw_result[sa_i]+sa_x;
    End;
  End;
  {$ELSE}
  Procedure Post_Process_FFT_Array_to_Last;
  Asm
                push      eax
                push      ecx

                xor       eax,eax
                mov       ecx,Global.FFT.Length

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@01:     fld       qword ptr FFT_Array[eax*DOUBLE_SIZE]
                fmul      st,st
                fld       qword ptr FFT_Array[eax*DOUBLE_SIZE+DOUBLE_SIZE]
                fmul      st,st
                fadd

                fst       qword ptr fft_last_result[eax*(DOUBLE_SIZE/2)]
                fadd      qword ptr fft_raw_result[eax*(DOUBLE_SIZE/2)]
                fstp      qword ptr fft_raw_result[eax*(DOUBLE_SIZE/2)]

                add       eax,2
                cmp       eax,ecx
                jle       @@01

                pop       ecx
                pop       eax
//                mov     Global.jump_location,offset @@01
  End;
  {$ENDIF}

  {$IFNDEF USEASM}
  //----------------------------------------------------------------------------
  Procedure Post_Process_FFT_Results();
  //----------------------------------------------------------------------------
  Var
    sc_i : Integer;
    sc_x : Extended;
  Begin
    for sc_i:=0 to Global.FFT.length_Half do
    begin
      with FFT_Array.DComplex[sc_i] do
        sc_x:=Re*Re+Im*Im;
      fft_last_result[sc_i]:=sc_x;
      fft_raw_result[sc_i]:=fft_raw_result[sc_i]+sc_x;
      fft_skew_result[sc_i]:=nroot(sc_x)*Skewing_Gain[sc_i shl Global.FFT.bit_shift_from_max];
    end
  //----------------------------------------------------------------------------
  End;
  //----------------------------------------------------------------------------
  {$ELSE}
  //============================================================================
  Procedure Post_Process_FFT_Results(); Assembler;
  //============================================================================
  Asm
                push    eax
                push    ebx
                push    ecx
                push    edx
  //============================================================================
  // Calculate magnitude of FFT_Array values and skew. Half FFT;
  //============================================================================
                mov     ecx,Global.FFT.bit_shift_from_max
                mov     edx,Global.FFT.Length
                mov     eax,0

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}

  //============================================================================
      @@01:       fld     qword ptr fft_array[eax*DOUBLE_SIZE]
                  fmul    st,st
                  fld     qword ptr fft_array[eax*DOUBLE_SIZE+DOUBLE_SIZE]
                  fmul    st,st
                  fadd

                  fld     st
                  fst     qword ptr fft_last_result[eax*(DOUBLE_SIZE/2)]
                  fadd    qword ptr fft_raw_result[eax*(DOUBLE_SIZE/2)]
                  fstp    qword ptr fft_raw_result[eax*(DOUBLE_SIZE/2)]

                  call    nroot_int

                  mov     ebx,eax
                  shl     ebx,cl
                  fmul    qword ptr Skewing_Gain[ebx*(DOUBLE_SIZE/2)]
                  fstp    qword ptr fft_skew_result[eax*(DOUBLE_SIZE/2)]

                  add     eax,2
                  cmp     eax,edx
                  jle     @@01

  //============================================================================
                pop     edx
                pop     ecx
                pop     ebx
                pop     eax
//                mov     Global.jump_location,offset @@01
  //============================================================================
  End;
  //============================================================================
  {$ENDIF}

  {$IFNDEF USEASM}
  //----------------------------------------------------------------------------
  Procedure Post_Process_FFT_Results_No_Raw();
  //----------------------------------------------------------------------------
  Var
    sc_i : Integer;
    sc_x : Extended;
  Begin
    for sc_i:=Spreading.Bins.Lo[Global.FFT.Bit_Length] to Spreading.Bins.Hi[Global.FFT.Bit_Length]+1 do
    begin
      with FFT_Array.DComplex[sc_i] do
        sc_x:=Re*Re+Im*Im;
      fft_skew_result[sc_i]:=nroot(sc_x)*Skewing_Gain[sc_i shl Global.FFT.bit_shift_from_max];
    end
  //----------------------------------------------------------------------------
  End;
  //----------------------------------------------------------------------------
  {$ELSE}
  //============================================================================
  Procedure Post_Process_FFT_Results_No_Raw(); Assembler;
  //============================================================================
  Asm
                push    eax
                push    ebx
                push    ecx
                push    edx
  //============================================================================
  // Calculate magnitude of FFT_Array values and skew. Half FFT;
  //============================================================================
                mov     ecx,Global.FFT.bit_shift_from_max
                mov     edx,Global.FFT.Bit_Length
                mov     eax,dword ptr Spreading.Bins.Lo[edx*integer_size]
                Mov     edx,dword ptr Spreading.Bins.Hi[edx*Integer_Size]
                add     edx,1
                add     edx,edx
                add     eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$ENDIF}

  //============================================================================
      @@01:       fld     qword ptr fft_array[eax*DOUBLE_SIZE]
                  fmul    st,st
                  fld     qword ptr fft_array[eax*DOUBLE_SIZE+DOUBLE_SIZE]
                  fmul    st,st
                  fadd

                  call    nroot_int

                  mov     ebx,eax
                  shl     ebx,cl
                  fmul    qword ptr Skewing_Gain[ebx*(DOUBLE_SIZE/2)]
                  fstp    qword ptr fft_skew_result[eax*(DOUBLE_SIZE/2)]

                  add     eax,2
                  cmp     eax,edx
                  jle     @@01

  //============================================================================
                pop     edx
                pop     ecx
                pop     ebx
                pop     eax
//                mov     Global.jump_location,offset @@01
  //============================================================================
  End;
  //============================================================================
  {$ENDIF}


  {$IFNDEF USEASM}
  //============================================================================
  Procedure Add_to_Histogram;
  //============================================================================
  Var
    sa_i, sa_j : Integer;
  Begin
  //============================================================================
    For sa_j:=0 to Global.Channels-1 do
      for sa_i:=0 to BlockData.Size.This-1 do
      Begin
        inc(History.Histogram_DATA[round(BlockData.Pointer[2]^[sa_j,sa_i]*History.Histogram_Multiplier)+History.Histogram_Offset]);
        inc(History.Histogram_BTRD[round(BlockData.BTRD[sa_j,sa_i]*History.Histogram_Multiplier)+History.Histogram_Offset]);
        inc(History.Histogram_CORR[round(BlockData.CORR[sa_j,sa_i]*History.Histogram_Multiplier)+History.Histogram_Offset]);
      End;
  //============================================================================
  End;
  //============================================================================
  {$ELSE}
  //============================================================================
  Procedure Add_to_Histogram;
  //============================================================================
  Var
    sa_j : Integer;
  Begin
  //============================================================================
    For sa_j:=0 to Global.Channels-1 do
    asm
                push    ebx
                push    edi
                push    edx

                mov     edi,sa_j
                xor     eax,eax

                fild    History.Histogram_Offset
                fld     History.Histogram_Multiplier

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;
                {$ENDIF}

    @@00:       mov     ecx,dword ptr BlockData.Channel_Offset[edi*INTEGER_SIZE]
                mov     ebx,dword ptr BlockData.Pointer[2*POINTER_SIZE]
                add     ebx,ecx

    @@01:       fild    dword ptr [ebx+eax*INTEGER_SIZE]
                fmul    st,st(1)
                fadd    st,st(2)
                fistp   dword ptr [esp]
                mov     edx,[esp]
                inc     dword ptr History.Histogram_DATA[edx*INTEGER_SIZE]

                fild    dword ptr BlockData.BTRD[ecx+eax*INTEGER_SIZE]
                fmul    st,st(1)
                fadd    st,st(2)
                fistp   dword ptr [esp]
                mov     edx,[esp]
                inc     dword ptr History.Histogram_BTRD[edx*INTEGER_SIZE]

                fild    dword ptr BlockData.CORR[ecx+eax*INTEGER_SIZE]
                fmul    st,st(1)
                fadd    st,st(2)
                fistp   dword ptr [esp]
                mov     edx,[esp]
                inc     dword ptr History.Histogram_CORR[edx*INTEGER_SIZE]

                add     eax,1
                cmp     eax,BlockData.Size.This
                jl      @@01

                fstp    st
                fstp    st

                pop     edx
                pop     edi
                pop     ebx
//                mov     Global.jump_location,offset @@00
    End;
  //============================================================================
  End;
  //============================================================================
  {$ENDIF}
  //============================================================================
  procedure process_this_codec_block;
  //============================================================================
  var
    this_channel                          : Integer;
    this_Analysis_Number                  : Integer;
    this_analysis_block_number            : Integer;
    this_analysis_block_start             : Integer;

    codec_block_dependent_bits_to_remove  : Integer;
    min_neg_codec_block_start             : Integer;
    max_pos_codec_block_start             : Integer;
    local_channels                        : Integer;

    Spreading_Old                         : Double;
    Spreading_New                         : Double;
    Spreading_result                      : Double;

    bits_removed_this_codec_block         : Double;
    this_fft_underlap_length              : Double;
    temp_result                           : Double;

    sa_i                                  : Integer;
    sa_j                                  : Integer;
    sa_k                                  : Integer;
    temp_btr                              : Integer;
    Spreading_Used                        : Integer;

    Fill_FFT_Proc                         : Function : Extended;
    Add_to_LongDist_Proc                  : Procedure;

    channel_data                          : Array[0..MAX_CHANNELS-1] of channel_data_rec;

    //==========================================================================
    Procedure Set_FFT_Variables;
    //==========================================================================
    Begin
    //==========================================================================
      if not FFTW_Initialised then
        if SetFFTLength(Global.FFT.Length_Half)=0 then
          lossyWAVerror('Fatal FFT error!',$88);

      min_neg_codec_block_start:=-(BlockData.Size.last+BlockData.Size.prev);
      max_pos_codec_block_start:=BlockData.Size.this+BlockData.Size.next-Global.FFT.length;

      this_analysis_block_start:=process.actual_analysis_blocks_start[Global.FFT.Bit_Length];
      this_fft_underlap_length:=process.fft_underlap_length[Global.FFT.Bit_Length];
    //==========================================================================
    End;
    //==========================================================================

    //==========================================================================
    Procedure Process_Data_to_LongDist;
    //==========================================================================
    Var
      this_analysis_block_number : Integer;
    Begin
    //==========================================================================
      Zero_FFT_Raw_Results;

      Process.Analyses_Performed:=0;
      for this_analysis_block_number:=0 to process.analysis_blocks[Global.FFT.Bit_Length]-1 do
      begin
        FillFFT.actual_analysis_block_start:=floor(this_analysis_block_start+this_analysis_block_number*this_fft_underlap_length);
        FillFFT.actual_analysis_block_start:=min(max(FillFFT.actual_analysis_block_start,min_neg_codec_block_start),max_pos_codec_block_start);

        inc(Process.Analyses_Performed);
        if Fill_FFT_Proc<>0 then
        begin
          if FFTW_Initialised then
            FFTW.EXECUTE(FFTW.plans[Global.FFT.Bit_Length])
          else
            FFT_DIT_Real();

          Post_Process_FFT_Array_to_Last;
        end;
      end;
      Add_to_LongDist_Proc;
    //==========================================================================
    End;
    //==========================================================================

  //============================================================================
  begin
  //============================================================================
    codec_block_dependent_bits_to_remove:=Global.Bits_Per_Sample;
    bits_removed_this_codec_block:=0;

    if parameters.midside and (Global.channels=2) then
      local_channels:=4
    else
      local_channels:=Global.channels;

    //==========================================================================
    for this_channel:=0 to local_channels-1 do
    //==========================================================================
    begin
      Global.channel:=this_channel;

      channel_data[Global.channel].maximum_bits_to_remove:=Settings.static_maximum_bits_to_remove;
      channel_data[Global.Channel].min_fft_result.btr:=channel_data[Global.channel].maximum_bits_to_remove;
      Channel_Data[Global.Channel].min_fft_result.analysis:=7;

      if parameters.dynamic=0 then
        Process.Dynamic_Maximum_Bits_to_Remove:=Settings.Static_Maximum_Bits_to_Remove
      else
        Process.Dynamic_Maximum_Bits_to_Remove:=Floor(max(0,floor(BlockData.Channel_Log2_RMS[Global.channel]-Settings.dynamic_minimum_bits_to_keep)));

      if Process.Dynamic_Maximum_Bits_to_Remove<Channel_Data[Global.Channel].maximum_bits_to_remove then
      begin
        channel_data[Global.channel].maximum_bits_to_remove:=Process.Dynamic_Maximum_Bits_to_Remove;
        channel_data[Global.Channel].min_fft_result.btr:=Process.Dynamic_Maximum_Bits_to_Remove;
        Channel_Data[Global.Channel].min_fft_result.analysis:=8;
      end;

      for this_Analysis_Number:=1 to PRECALC_ANALYSES do
      begin
        Global.Analysis_Number:=this_Analysis_Number;
        Set_FFT_Data_Rec(Global.FFT,Global.Analysis.Bits[Global.Analysis_Number]);

        Results.This_FFT.btr:=99;
        Results.This_FFT.spreading:=db_MAXdB;
        Results.This_FFT.analysis:=Global.Analysis_Number;

        if ((Settings.fft_analysis_switches and Global.FFT.length)>0) and (Global.FFT.length<=Global.codec_block_size*2)then
        begin
          Set_FFT_Variables;

          Zero_FFT_Raw_Results;

          process.analyses_performed:=0;
          for this_analysis_block_number:=0 to process.analysis_blocks[Global.FFT.Bit_Length] do
          begin
            FillFFT.actual_analysis_block_start:=floor(this_analysis_block_start+this_analysis_block_number*this_fft_underlap_length);
            FillFFT.actual_analysis_block_start:=min(max(FillFFT.actual_analysis_block_start,min_neg_codec_block_start),max_pos_codec_block_start);

            if (this_analysis_block_number=0) and (FillFFT.Actual_Analysis_Block_Start=Results.Saved_FFT_Results[Global.Channel,Global.Analysis_Number].start) then
            Begin
              if (Global.FFT.Length and Settings.raw_result_short)>0 then
                Fill_Last_From_History;

              if (Global.FFT.Length and Settings.raw_result_sgns)>0 then
                Fill_Last_from_SGNS;

              Spreading.Old_Minimum:=Process.FFT_Spreading[Global.Analysis_Number,Global.Channel].Old_Minimum;
              Spreading.New_Minimum:=Process.FFT_Spreading[Global.Analysis_Number,Global.Channel].New_Minimum;
              Spreading.Alt_Average:=Process.FFT_Spreading[Global.Analysis_Number,Global.Channel].Alt_Average;
            End
            else
              if FillFFT_Input_From_Data()=0 then
              Begin
                if (Global.FFT.Length and (Settings.raw_result_short or Settings.raw_result_sgns))>0 then
                  Fill_Last_With_Zero;
                Spreading.Old_Minimum:=db_MAXdB;
                Spreading.New_Minimum:=db_MAXdB;
                Spreading.Alt_Average:=db_MAXdB;
              End
              else
              begin
                if FFTW_Initialised then
                  FFTW.EXECUTE(FFTW.plans[Global.FFT.Bit_Length])
                else
                  FFT_DIT_Real();

                Process.Post_Analysis[Global.Analysis_Number];

                Spreading_Function();
              end;

            Inc(Process.Analyses_Performed);

            Spreading_Result:=Spreading.Alt_Average;
            Spreading_Used:=1;
            If Spreading.New_Minimum<Spreading_Result then
            Begin
              Spreading_Result:=Spreading.New_Minimum;
              Spreading_Used:=2;
            End;
            If Spreading.Old_Minimum<Spreading_Result then
            Begin
              Spreading_Result:=Spreading.Old_Minimum;
              Spreading_Used:=3;
            End;

            Results.This_FFT.Spreading:=max(0,Spreading_result-FillFFT.window_function_RMS);
            Results.This_FFT.btr:=Spreading.threshold_index[floor(Min(THRESHOLD_INDEX_SPREAD_MUL-1,Results.This_FFT.spreading)*THRESHOLD_INDEX_SPREAD)];
            Results.This_FFT.start:=FillFFT.actual_analysis_block_start-Global.codec_block_size;
            Results.This_FFT.analysis:=Global.Analysis_Number;

            if Parameters.Spread<>-1 then
            Begin
              Case Spreading_Used of
                1 : Inc(Process.Alt_Ave_Used[Global.Analysis_Number]);
                2 : Begin
                      Inc(Process.New_Min_Used_History[Global.Analysis_Number,Process.New_Min_Bin]);
                      Inc(Process.New_Min_Used[Global.Analysis_Number]);
                    End;
                3 : Begin
                      Inc(Process.Old_Min_Used_History[Global.Analysis_Number,Process.Old_Min_Bin]);
                      Inc(Process.Old_Min_Used[Global.Analysis_Number]);
                    End;
              End;

              if Results.This_FFT.BTR>Settings.Static_Maximum_Bits_to_Remove then
                inc(Process.Over_Static[Global.Analysis_Number]);

              if Results.This_FFT.BTR>Process.Dynamic_Maximum_Bits_to_Remove then
                inc(Process.Over_Dynamic[Global.Analysis_Number]);
            end;

            if (Results.This_FFT.btr<channel_data[Global.channel].min_fft_result.btr) or (channel_data[Global.channel].min_fft_result.btr=-1) then
              channel_data[Global.channel].min_fft_result:=Results.This_FFT;
          end;

          Process.Analyses_Completed[Global.Analysis_Number]:=Process.Analyses_Completed[Global.Analysis_Number]+Process.Analyses_Performed;

          if ((Global.FFT.Length and settings.raw_result_short)>0) then
            Fill_History_from_Last;

          if ((Global.FFT.Length and settings.raw_result_SGNS)>0) then
            Fill_SGNSFFT_from_Last;

          Results.saved_fft_results[Global.Channel,Global.Analysis_Number]:=Results.This_FFT;

          Process.FFT_Spreading[Global.Analysis_Number,Global.Channel].Old_Minimum:=Spreading.Old_Minimum;
          Process.FFT_Spreading[Global.Analysis_Number,Global.Channel].New_Minimum:=Spreading.New_Minimum;
          Process.FFT_Spreading[Global.Analysis_Number,Global.Channel].Alt_Average:=Spreading.Alt_Average;
        end;
      end;

      If (Parameters.Adaptive_Shaping<>-1) then
        Make_Filter;

      if channel_data[Global.channel].min_fft_result.btr<0 then
        channel_data[Global.channel].min_fft_result.btr:=0;

      channel_data[Global.channel].bits_to_remove:=channel_data[Global.channel].min_fft_result.btr;

      if Global.channel<Global.channels then
      begin
        remove_bits_data.bits_to_remove:=channel_data[Global.channel].bits_to_remove;
        remove_bits;

        channel_data[Global.channel].bits_to_remove:=remove_bits_data.bits_removed;
        channel_data[Global.channel].clips:=remove_bits_data.clips;
        codec_block_dependent_bits_to_remove:=min(codec_block_dependent_bits_to_remove,channel_data[Global.channel].bits_to_remove);
      end;
    //==========================================================================
    end;
    //==========================================================================

    //==========================================================================
    for this_channel:=0 to Global.channels-1 do
    //==========================================================================
    begin
      Global.channel:=this_channel;

      inc(results.minima[Global.Channel,Channel_Data[Global.Channel].min_fft_result.analysis]);

      if Parameters.linkchannels or parameters.midside then
      begin
        remove_bits_data.bits_to_remove:=codec_block_dependent_bits_to_remove;
        if channel_data[Global.channel].bits_to_remove<>remove_bits_data.bits_to_remove then
        begin
          if Global.channel<Global.channels then
            remove_bits();

          channel_data[Global.channel].bits_to_remove:=remove_bits_data.bits_removed;
          channel_data[Global.channel].clips:=remove_bits_data.clips;
        end;
      end;

      channel_data[Global.channel].bits_lost:=channel_data[Global.channel].min_fft_result.btr-channel_data[Global.channel].bits_to_remove;

      inc(Stats.Bits_Removed[Global.Channel,Channel_Data[Global.Channel].bits_to_remove]);

      Global.bits_removed:=Global.bits_removed+channel_data[Global.channel].bits_to_remove;
      Global.bits_lost:=Global.bits_lost+channel_data[Global.channel].bits_lost;
      Global.clipped_samples:=Global.clipped_samples+channel_data[Global.channel].clips;

      if Parameters.detail then
        bit_removal_history^[(Global.blocks_processed-1)*Global.channels+Global.channel]:=channel_data[Global.channel].bits_to_remove;

      bits_removed_this_codec_block:=bits_removed_this_codec_block+OneOver[Global.channels]*channel_data[Global.channel].bits_to_remove;

    //==========================================================================
    end;
    //==========================================================================

    //==========================================================================
    // Post analyse bit removed / correction audio data using short FFT.
    //==========================================================================
    if Parameters.postanalyse then
    begin
    //==========================================================================
      Set_FFT_Data_Rec(Global.FFT,History.FFT.Bit_Length);

      Set_FFT_Variables;

      //========================================================================
      for this_channel:=0 to Global.channels-1 do
      Begin
      //========================================================================
        Global.Channel:=this_channel;

        //======================================================================
        // Create spectrum for BTRD-Codec-Block
        //======================================================================
        Fill_FFT_Proc:=@FillFFT_Input_From_BTRD;
        Add_to_LongDist_Proc:=@Add_to_BTRD_History;
        Process_DATA_to_Longdist;
        //======================================================================

        //======================================================================
        // Create spectrum for CORR-Codec-Block
        //======================================================================
        Fill_FFT_Proc:=@FillFFT_Input_From_CORR;
        Add_to_LongDist_Proc:=@Add_to_CORR_History;
        Process_DATA_to_Longdist;
        //======================================================================

      //========================================================================
      End;
      //========================================================================

    //==========================================================================
    end;
    //==========================================================================

    //==========================================================================
    // Post analyse DATA/BTRD/CORR audio data using Codec_Block_Length FFT.
    //==========================================================================
    if Parameters.longdist then
    begin
    //==========================================================================
      Set_FFT_Data_Rec(Global.FFT,LongDist.FFT.Bit_Length);

      Set_FFT_Variables;

      //========================================================================
      for this_channel:=0 to Global.channels-1 do
      Begin
      //========================================================================
        Global.Channel:=this_channel;

        //======================================================================
        // Create spectrum for DATA-Codec-Block
        //======================================================================
        Fill_FFT_Proc:=@FillFFT_Input_From_DATA;
        Add_to_LongDist_Proc:=@Add_to_DATA_LongDist;
        Process_DATA_to_Longdist;
        //======================================================================

        //======================================================================
        // Create spectrum for BTRD-Codec-Block
        //======================================================================
        Fill_FFT_Proc:=@FillFFT_Input_From_BTRD;
        Add_to_LongDist_Proc:=@Add_to_BTRD_LongDist;
        Process_DATA_to_Longdist;
        //======================================================================

        //======================================================================
        // Create spectrum for CORR-Codec-Block
        //======================================================================
        Fill_FFT_Proc:=@FillFFT_Input_From_CORR;
        Add_to_LongDist_Proc:=@Add_to_CORR_LongDist;
        Process_DATA_to_Longdist;
        //======================================================================

      //========================================================================
      End;
      //========================================================================

    //==========================================================================
    end;
    //==========================================================================

    //==========================================================================
    // Create DATA, BTRD and CORR sample value histogram
    //==========================================================================
    If Parameters.Histogram then
      Add_To_Histogram;
    //==========================================================================

    //==========================================================================
    // Store BTRD and CORR results, if necessary, for use next codec-block.
    //==========================================================================
    If Parameters.PostAnalyse or Parameters.LongDist then
      Fill_Audio_Safes;
    //==========================================================================

    Global.samples_processed:=(Global.blocks_processed-1)*Global.codec_block_size+BlockData.Size.this;

    Process_Output;

    //==========================================================================
    // Create histogram of LSBs and MSBs on a sample / block basis.
    //==========================================================================
    if Parameters.blockdist or Parameters.sampledist then
      lsb_analysis();
    //==========================================================================

  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure nProcess_Init;
  //============================================================================
  var
    sa_i, sa_j : Integer;
    total_overlap_length : Integer;
  Begin
  //============================================================================
    For sa_j:=1 to Precalc_Analyses do
    Begin
      for sa_i:=0 to MAX_FFT_LENGTH_HALF-1 do
      Begin
        Process.Old_Min_Used_History[sa_j,sa_i]:=0;
        Process.New_Min_Used_History[sa_j,sa_i]:=0;
      End;

      Process.Old_Min_Used[sa_j]:=0;
      Process.New_Min_Used[sa_j]:=0;
      Process.Alt_Ave_Used[sa_j]:=0;
      Process.Over_Dynamic[sa_j]:=0;
      Process.Over_Static[sa_j]:=0;

      Process.Post_Analysis[sa_j]:=@Post_Process_FFT_Results_No_Raw;
    End;

    if settings.raw_result_short<>0 then
      Process.Post_Analysis[2]:=@Post_Process_FFT_Results;

    if settings.raw_result_SGNS<>0 then
      Process.Post_Analysis[6]:=@Post_Process_FFT_Results;

    for sa_i:=1 to MAX_FFT_Bit_Length do
    begin
      Set_FFT_Data_Rec(Global.FFT,sa_i);

      if FFTW_Initialised then
        if ((settings.FFT_analysis_switches and Global.FFT.Length)>0) or (Parameters.LongDist and (Global.FFT.Bit_Length=LongDist.FFT.Bit_Length)) then
          FFTW.plans[Global.FFT.Bit_Length]:=FFTW.plan_dft_r2c_1d(Global.FFT.length,@FFT_Array,@FFT_Array,$69);

      //========================================================================
      // Calculate actual analysis_time values for FFT_lengths
      //========================================================================
      process.end_overlap_length[Global.FFT.Bit_Length]:=min(Global.FFT.length div 2,Global.codec_block_size);
      process.actual_analysis_blocks_start[Global.FFT.Bit_Length]:=-process.end_overlap_length[Global.FFT.Bit_Length];

      total_overlap_length:=Global.codec_block_size+process.end_overlap_length[Global.FFT.Bit_Length]*2-Global.FFT.length;
      process.FFT_underlap_length[Global.FFT.Bit_Length]:=max(1,Global.FFT.length div max(2,parameters.underlap));

      process.analysis_blocks[Global.FFT.Bit_Length]:=max(0,round(total_overlap_length/process.FFT_underlap_length[Global.FFT.Bit_Length]));

      process.FFT_underlap_length[Global.FFT.Bit_Length]:=total_overlap_length/max(1,process.analysis_blocks[Global.FFT.Bit_Length]);
      //========================================================================
    end;

  //============================================================================
  End;
  //============================================================================


//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2.Data[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
