//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nShiftBlocks;
//==============================================================================
interface
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================
Procedure Shift_Codec_Blocks();
//==============================================================================
implementation
//==============================================================================
uses
  nCore,
  nMaths,
  Math;
const
  UNITNAME='nShiftBlocks';
var
  Unit_Data_Packer2 : Packed Record
    Data : Array[0..2] of Integer;
  End;

//==============================================================================
 {$IFNDEF USEASM}
  //----------------------------------------------------------------------------
  Procedure Shift_Codec_Blocks();
  //----------------------------------------------------------------------------
  var
   sc_i : Integer;
   sc_p : Pointer;
   channel : Integer;
   left_sample, right_sample : Double;

    //--------------------------------------------------------------------------
    Function Channel_RMS(this_channel : Integer) : Extended;
    //--------------------------------------------------------------------------
    var
      Channel_Temp : Extended;
      sc_i : Integer;
    Begin
      channel_temp:=0;
      for sc_i:=0 to BlockData.Size.This-1 do
        channel_temp:=channel_temp+sqr(1.0*BlockData.Pointer[2]^[this_channel,sc_i]);
      channel_RMS:=nlog2(channel_temp*Global.Codec_Block_Size_Recip)*SINGLE_HALF
    //--------------------------------------------------------------------------
    End;
    //--------------------------------------------------------------------------

  begin
    sc_p:=BlockData.Pointer[0];
    BlockData.Pointer[0]:=BlockData.Pointer[1];
    BlockData.Pointer[1]:=BlockData.Pointer[2];
    BlockData.Pointer[2]:=BlockData.Pointer[3];
    BlockData.Pointer[3]:=sc_p;

    BlockData.Size.Prev:=BlockData.Size.Last;
    BlockData.Size.Last:=BlockData.Size.This;
    BlockData.Size.This:=BlockData.Size.Next;
    BlockData.Size.Next:=0;

    if BlockData.Size.This>0 then
    begin
      Global.Codec_block_size_recip:=OneOver[BlockData.Size.This];
      inc(Global.blocks_processed);
      Global.blocks_processed_recip:=1/Global.blocks_processed;

      for channel:=0 to Global.channels-1 do
        BlockData.Channel_Log2_RMS[channel]:=Channel_RMS(channel);

      //========================================================================
      // optional examination of mid-side content.
      //========================================================================
      if Parameters.midside and (Global.channels=2) then
      begin
        for sc_i:=0 to BlockData.Size.This-1 do
        begin
          left_sample:=BlockData.Pointer[2]^[0,sc_i];
          right_sample:=BlockData.Pointer[2]^[1,sc_i];

          BlockData.Pointer[2]^[2,sc_i]:=round(single_half*(left_sample+right_sample));
          BlockData.Pointer[2]^[3,sc_i]:=round(single_half*(left_sample-right_sample));
        end;

        BlockData.Channel_Log2_RMS[2]:=Channel_RMS(2);
        BlockData.Channel_Log2_RMS[3]:=Channel_RMS(3);
      end;
    end
    else
      for channel:=0 to Global.channels-1 do
        BlockData.Channel_Log2_RMS[channel]:=0;
  end;
  //----------------------------------------------------------------------------
  {$ELSE}
  //============================================================================
  Procedure Shift_Codec_Blocks();
  //============================================================================
  asm
                push    ebp
                push    ebx
                push    edi
                push    esi

                //==============================================================
                // Shift Data Block Pointers
                //==============================================================
                mov     eax,dword ptr BlockData.Pointer[3*POINTER_SIZE]
                xchg    eax,dword ptr BlockData.Pointer[2*POINTER_SIZE]
                xchg    eax,dword ptr BlockData.Pointer[1*POINTER_SIZE]
                xchg    eax,dword ptr BlockData.Pointer[0*POINTER_SIZE]
                mov     dword ptr BlockData.Pointer[3*POINTER_SIZE],eax
                //==============================================================

                //==============================================================
                // Shift Data Block Sample Sizes
                //==============================================================
                xor     eax,eax
                xchg    eax,BlockData.Size.next
                xchg    eax,BlockData.Size.this
                xchg    eax,BlockData.Size.last
                mov     BlockData.Size.prev,eax
                //==============================================================

                mov     eax,BlockData.Size.this
                cmp     eax,1
                jl      @@99

                fld     qword ptr OneOver[eax*DOUBLE_SIZE]
                fstp    Global.codec_block_size_recip

                add     Global.blocks_processed,1
                fld1
                fidiv   Global.blocks_processed
                fstp    Global.blocks_processed_recip

                mov     ebx,Global.channels
                mov     ebp,dword ptr BlockData.Pointer[2*POINTER_SIZE]

                jmp     @@04

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN32}
                {$ENDIF}
                {$ENDIF}
                {$ENDIF}

      @@00:       fldz
                  mov     ecx,dword ptr BlockData.Channel_Offset[ebx*POINTER_SIZE]
                  add     ecx,ebp
                  mov     edx,eax

                  {$IFDEF ALIGN08}
                  nop;nop;nop;
                  {$IFDEF ALIGN16}
                  {$IFDEF ALIGN32}
                  nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                  {$ENDIF}
                  {$ENDIF}
                  {$ENDIF}

      @@01:       sub     edx,1
                  jl      @@03

                  fild    dword ptr [ecx+edx*INTEGER_SIZE]
                  fmul    st,st
                  fadd
                  jmp     @@01

                  {$IFDEF ALIGN08}
                  nop;nop;
                  {$IFDEF ALIGN16}
                  {$IFDEF ALIGN32}
                  {$ENDIF}
                  {$ENDIF}
                  {$ENDIF}

      @@03:     fmul    Global.codec_block_size_recip
                call    nlog2_int
                fmul    SINGLE_HALF
                fstp    qword ptr BlockData.Channel_Log2_RMS[ebx*DOUBLE_SIZE]

                {$IFDEF ALIGN08}
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN32}
                {$ENDIF}
                {$ENDIF}
                {$ENDIF}

      @@04:     sub     ebx,1
                jge     @@00

      @@05:     test    parameters.midside,1
                jz      @@99

                cmp     Global.channels,2
                jne     @@99

                fldz
                fldz

                mov     ecx,ebp
                add     ecx,dword ptr BlockData.Channel_Offset[1*INTEGER_SIZE]
                mov     edi,ebp
                add     edi,dword ptr BlockData.Channel_Offset[2*INTEGER_SIZE]
                mov     esi,ebp
                add     esi,dword ptr BlockData.Channel_Offset[3*INTEGER_SIZE]

                mov     eax,BlockData.Size.this
                jmp     @@07

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN32}
                nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}
                {$ENDIF}

      @@06:     fild    dword ptr [ebp+eax*INTEGER_SIZE]
                fiadd   dword ptr [ecx+eax*INTEGER_SIZE]
                fmul    SINGLE_HALF
                frndint
                fist    dword ptr [edi+eax*INTEGER_SIZE]
                fmul    st,st
                fadd

                fild    dword ptr [ebp+eax*INTEGER_SIZE]
                fisub   dword ptr [ecx+eax*INTEGER_SIZE]
                fmul    SINGLE_HALF
                frndint
                fist    dword ptr [esi+eax*INTEGER_SIZE]
                fmul    st,st
                fadd    st,st(2)
                fstp    st(2)

      @@07:     sub     eax,1
                jge     @@06

                fmul    Global.codec_block_size_recip
                call    nlog2_int
                fmul    SINGLE_HALF
                fstp    qword ptr BlockData.Channel_Log2_RMS[2*double_size]

                fmul    Global.codec_block_size_recip
                call    nlog2_int
                fmul    SINGLE_HALF
                fstp    qword ptr BlockData.Channel_Log2_RMS[3*double_size]

      @@99:     //fwait
                pop     esi
                pop     edi
                pop     ebx
                pop     ebp
  end;
  //============================================================================
  {$ENDIF}

 //==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2.Data[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
